# Cherrypy-foundation
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cherrypy
from cherrypy.test import helper

from cherrypy_foundation.tools import secure_headers  # noqa


@cherrypy.tools.secure_headers(
    csp={
        "default-src": "'self'",
        "style-src": ("'self'", "'unsafe-inline'"),
        "script-src": ("'self'", "'unsafe-inline'"),
        "img-src": ("'self'", "data:"),
    }
)
@cherrypy.tools.proxy(local=None, remote='X-Real-IP')
class Root:

    @cherrypy.expose
    @cherrypy.tools.sessions(locking='explicit')
    def index(self):
        return {'var1': 'test-jinja2'}

    @cherrypy.expose
    @cherrypy.tools.sessions(locking='explicit')
    def logout(self, **kwargs):
        raise cherrypy.HTTPRedirect('/')

    @cherrypy.expose
    @cherrypy.tools.secure_headers(on=False)
    @cherrypy.tools.sessions(on=False)
    def static(self, value=None):
        return 'OK'


class SecureHeaderTest(helper.CPWebCase):
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.tree.mount(Root(), '/')

    def test_cookie_samesite_lax(self):
        # Given a request made to rdiffweb
        # When receiving the response
        self.getPage('/')
        # Then the header contains Set-Cookie with SameSite=Lax
        cookie = self.assertHeader('Set-Cookie')
        self.assertIn('SameSite=Lax', cookie)

    def test_cookie_samesite_lax_without_session(self):
        # Given not a client sending no cookie
        self.cookies = None
        # When a query is made to a static path (without session)
        self.getPage('/static/blue.css')
        # Then Set-Cookie is not defined.
        self.assertNoHeader('Set-Cookie')

    def test_cookie_with_https(self):
        # Given an https request made to rdiffweb
        self.getPage('/', headers=[('X-Forwarded-Proto', 'https')])
        # When receiving the response
        self.assertStatus(200)
        # Then the header contains Set-Cookie with Secure
        cookie = self.assertHeader('Set-Cookie')
        self.assertIn('Secure', cookie)

    def test_cookie_with_http(self):
        # Given an https request made to rdiffweb
        self.getPage('/')
        # When receiving the response
        # Then the header contains Set-Cookie with Secure
        cookie = self.assertHeader('Set-Cookie')
        self.assertNotIn('Secure', cookie)

    def test_get_with_wrong_origin(self):
        # Given a GET request made to rdiffweb
        # When the request is made using a different origin
        self.getPage('/', headers=[('Origin', 'http://www.examples.com')])
        # Then the response status it 200 OK.
        self.assertStatus(200)

    def test_post_with_wrong_origin(self):
        # Given a POST request made to rdiffweb
        # When the request is made using a different origin
        self.getPage('/logout', headers=[('Origin', 'http://www.examples.com')], method='POST')
        # Then the request is refused with 403 Forbiden
        self.assertStatus(403)
        self.assertInBody('Unexpected Origin header')

    def test_post_with_prefixed_origin(self):
        # Given a POST request made to rdiffweb
        # When the request is made using a different origin
        base = 'http://%s:%s' % (self.HOST + 'anything.com', self.PORT)
        self.getPage('/logout', headers=[('Origin', base)], method='POST')
        # Then the request is accepted with 200 OK
        self.assertStatus(403)
        self.assertInBody('Unexpected Origin header')

    def test_post_with_valid_origin(self):
        # Given a POST request made to rdiffweb
        # When the request is made using a different origin
        base = 'http://%s:%s' % (self.HOST, self.PORT)
        self.getPage('/logout', headers=[('Origin', base)], method='POST')
        # Then the request is accepted with 200 OK
        self.assertStatus(303)

    def test_post_without_origin(self):
        # Given a POST request made to rdiffweb
        # When the request is made without an origin
        self.getPage('/logout', method='POST')
        # Then the request is accepted with 200 OK
        self.assertStatus(303)

    def test_clickjacking_defense(self):
        # Given a POST request made to rdiffweb
        # When the request is made without an origin
        self.getPage('/')
        # Then the request is accepted with 200 OK
        self.assertStatus(200)
        self.assertHeaderItemValue('X-Frame-Options', 'DENY')

    def test_no_cache(self):
        # Given a POST request made to rdiffweb
        # When the request is made without an origin
        self.getPage('/')
        # Then the request is accepted with 200 OK
        self.assertStatus(200)
        self.assertHeaderItemValue('Cache-control', 'no-cache')
        self.assertHeaderItemValue('Cache-control', 'no-store')
        self.assertHeaderItemValue('Cache-control', 'must-revalidate')
        self.assertHeaderItemValue('Cache-control', 'max-age=0')
        self.assertHeaderItemValue('Pragma', 'no-cache')
        self.assertHeaderItemValue('Expires', '0')

    def test_no_cache_with_static(self):
        self.getPage('/static')
        # Then the request is accepted with 200 OK
        self.assertStatus(200)
        self.assertNoHeader('Cache-control')
        self.assertNoHeader('Pragma')
        self.assertNoHeader('Expires')

    def test_referrer_policy(self):
        # Given a POST request made to rdiffweb
        # When the request is made without an origin
        self.getPage('/')
        # Then the request is accepted with 200 OK
        self.assertStatus(200)
        self.assertHeaderItemValue('Referrer-Policy', 'same-origin')

    def test_nosniff(self):
        # Given a POST request made to rdiffweb
        # When the request is made without an origin
        self.getPage('/')
        # Then the request is accepted with 200 OK
        self.assertStatus(200)
        self.assertHeaderItemValue('X-Content-Type-Options', 'nosniff')

    def test_xss_protection(self):
        # Given a POST request made to rdiffweb
        # When the request is made without an origin
        self.getPage('/')
        # Then the request is accepted with 200 OK
        self.assertStatus(200)
        self.assertHeaderItemValue('X-XSS-Protection', '1; mode=block')

    def test_content_security_policy(self):
        # Given a POST request made to rdiffweb
        # When the request is made without an origin
        self.getPage('/')
        # Then the request is accepted with 200 OK
        self.assertStatus(200)
        self.assertHeaderItemValue(
            'Content-Security-Policy',
            "default-src 'self'; style-src 'self' 'unsafe-inline'; script-src 'self' 'unsafe-inline'; img-src 'self' data:",
        )

    def test_strict_transport_security(self):
        # Given a POST request made to rdiffweb
        # When the request is made without an origin
        self.getPage('/', headers=[('X-Forwarded-Proto', 'https')])
        # Then the request is accepted with 200 OK
        self.assertStatus(200)
        self.assertHeaderItemValue('Strict-Transport-Security', 'max-age=31536000; includeSubDomains')



# MFA tool for cherrypy
# Copyright (C) 2022-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import datetime
import logging
import secrets
import string
from typing import Optional

import cherrypy

from cherrypy_foundation.sessions import session_lock

from ..passwd import check_password, hash_password

MFA_CODE = '_auth_mfa_code'
MFA_CODE_TIME = '_auth_mfa_code_time'
MFA_CODE_ATTEMPT = '_auth_mfa_code_attempt'
MFA_REDIRECT_URL = '_auth_mfa_redirect_url'
MFA_TRUSTED_IP_LIST = '_auth_mfa_trusted_ip_list'
MFA_USER_KEY = '_auth_mfa_user_key'
MFA_VERIFICATION_TIME = '_auth_mfa_time'

MFA_DEFAULT_CODE_TIMEOUT = 10  # minutes
MFA_DEFAULT_LENGTH = 8
MFA_DEFAULT_MAX_ATTEMPT = 3
MFA_DEFAULT_MAX_TRUSTED_IPS = 5
MFA_DEFAULT_TRUST_DURATION = 43200  # 30 days


class CheckAuthMfa(cherrypy.Tool):
    def __init__(self, priority: int = 75):
        super().__init__(point='before_handler', callable=self.run, priority=priority)

    # ---- Config helpers ----

    @staticmethod
    def _code_length() -> int:
        """Return the configured code length."""
        length = cherrypy.request.config.get('tools.auth_mfa.code_length', MFA_DEFAULT_LENGTH)
        return max(1, int(length))

    @staticmethod
    def _code_timeout_minutes() -> int:
        # Lifetime for a one-time MFA code
        return int(cherrypy.request.config.get('tools.auth_mfa.code_timeout', MFA_DEFAULT_CODE_TIMEOUT))

    @staticmethod
    def _max_attempts() -> int:
        return int(cherrypy.request.config.get('tools.auth_mfa.max_attempts', MFA_DEFAULT_MAX_ATTEMPT))

    @staticmethod
    def _max_trusted_ips() -> int:
        return int(cherrypy.request.config.get('tools.auth_mfa.max_trusted_ips', MFA_DEFAULT_MAX_TRUSTED_IPS))

    @staticmethod
    def _trust_duration_minutes() -> int:
        # How long a device/IP remains “MFA-verified”
        return int(cherrypy.request.config.get('tools.auth_mfa.trust_duration', MFA_DEFAULT_TRUST_DURATION))

    # ---- Core operations ----

    def generate_code(self) -> str:
        """
        Generate a random numeric code, store its hash and metadata in the session,
        and return the cleartext so the caller can deliver it out-of-band.
        """
        if not hasattr(cherrypy.serving, 'session'):
            raise cherrypy.HTTPError(500, 'MFA requires sessions')

        if not cherrypy.request.login:
            raise cherrypy.HTTPError(401, 'MFA requires an authenticated username')

        length = self._code_length()
        code = ''.join(secrets.choice(string.digits) for _ in range(length))
        with session_lock() as session:
            session[MFA_USER_KEY] = cherrypy.request.login
            session[MFA_CODE] = hash_password(code)
            session[MFA_CODE_TIME] = session.now()
            session[MFA_CODE_ATTEMPT] = 0
        return code

    def _is_verified(self) -> bool:
        """Return True if current user/IP is within the MFA trust window."""
        if not cherrypy.request.login or not hasattr(cherrypy.serving, 'session'):
            return False

        # Check if our username match current login user.
        with session_lock() as session:
            if session.get(MFA_USER_KEY) != cherrypy.request.login:
                return False

            # Check if the current IP is trusted
            ip_list = session.get(MFA_TRUSTED_IP_LIST) or []
            if cherrypy.request.remote.ip not in ip_list:
                return False

            # Check if user ever pass the verification.
            verified_at: Optional[datetime.datetime] = session.get(MFA_VERIFICATION_TIME)
            if not verified_at:
                return False

            # Check trusted duration time
            trust_minutes = self._trust_duration_minutes()
            if (verified_at + datetime.timedelta(minutes=trust_minutes)) <= session.now():
                return False

        return True

    def is_code_expired(self) -> bool:
        """
        Return True if the current user's MFA code is absent, expired, or attempts exceeded.
        """
        # Check session existence first to avoid AttributeError
        if not hasattr(cherrypy.serving, 'session'):
            return True

        with session_lock() as session:
            if session.get(MFA_USER_KEY) != cherrypy.request.login:
                return True

            # Check if code is defined.
            hash_ = session.get(MFA_CODE)
            if not hash_:
                return True

            # Check issued time.
            issued_at: Optional[datetime.datetime] = session.get(MFA_CODE_TIME)
            if not issued_at or (
                (issued_at + datetime.timedelta(minutes=self._code_timeout_minutes())) < session.now()
            ):
                return True

            # Check number of attempt.
            attempts = int(session.get(MFA_CODE_ATTEMPT, 0))
            if attempts >= self._max_attempts():
                return True

        return False

    # ---- Tool entrypoint ----

    def run(self, mfa_url: str = '/mfa/', mfa_enabled=True, **_):
        """
        Gate requests for users who have MFA enabled but are not yet verified.
        mfa_enabled can be a bool or a callable returning bool.
        """
        enabled = mfa_enabled() if callable(mfa_enabled) else bool(mfa_enabled)

        # Normalize request path
        request = cherrypy.serving.request
        req_path = request.path_info or '/'
        mfa_path = mfa_url if mfa_url.startswith('/') else '/' + mfa_url

        # If the current request is the MFA page itself:
        if req_path.rstrip('/') == mfa_path.rstrip('/'):
            # If disabled or already verified, send user back to the original URL (or home)
            if not enabled or self._is_verified():
                raise cherrypy.tools.auth.redirect_to_original_url()
            return  # Allow the MFA page handler to render

        # If MFA is globally disabled for this user/realm, allow through
        if not enabled:
            return

        # Final gate: redirect to MFA if not verified
        if not self._is_verified():
            cherrypy.tools.auth.save_original_url()
            # Use a relative, safe redirect
            raise cherrypy.HTTPRedirect(mfa_path)

    # ---- Verification API for the MFA page handler ----

    def verify_code(self, code: str, persistent: bool = False) -> bool:
        """
        Verify the supplied one-time code. On success:
          - mark session persistent if requested
          - stamp verification time
          - add client IP to trusted list (dedup, cap)
          - clear the code, rotate session id
        """
        if not hasattr(cherrypy.serving, 'session'):
            return False

        with session_lock() as session:
            stored_hash = session.get(MFA_CODE)
            is_expired = self.is_code_expired()
            if self.is_code_expired():
                return False

            # Always perform the hash check regardless of expiration
            # to prevent timing attacks
            code_valid = False
            if stored_hash:
                code_valid = check_password(code, stored_hash)

            # Check all conditions after hash verification
            if is_expired or not code_valid:
                if not is_expired:  # Only increment if not expired
                    session[MFA_CODE_ATTEMPT] = attempts = int(session.get(MFA_CODE_ATTEMPT, 0)) + 1
                cherrypy.log(
                    f'verification failed user={cherrypy.request.login} ip={cherrypy.request.remote.ip} attempts={attempts}',
                    context='MFA',
                    severity=logging.WARNING,
                )
                return False

            # Success: trust this device/IP for the configured duration
            session[MFA_VERIFICATION_TIME] = session.now()
            ip_list = list({*(session.get(MFA_TRUSTED_IP_LIST) or []), cherrypy.request.remote.ip})
            # Cap the list to avoid unbounded growth (keep most recent N)
            max_ips = self._max_trusted_ips()
            if len(ip_list) > max_ips:
                ip_list = ip_list[-max_ips:]
            session[MFA_TRUSTED_IP_LIST] = ip_list

            # Clear the one-time code
            session[MFA_CODE] = None
            session[MFA_CODE_TIME] = None
            session[MFA_CODE_ATTEMPT] = 0

            # Honor “remember this device” option by making the session persistent
            if hasattr(cherrypy.tools, 'sessions_timeout'):
                try:
                    cherrypy.tools.sessions_timeout.set_persistent(bool(persistent))
                except Exception:
                    pass

            # Rotate session id to prevent fixation
            session.regenerate()
            cherrypy.log(
                f'verification successful user={cherrypy.request.login} ip={cherrypy.request.remote.ip}', context='MFA'
            )
        return True


cherrypy.tools.auth_mfa = CheckAuthMfa()



# LDAP Plugins for cherrypy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
Created on Oct 17, 2015

@author: Patrik Dufresne <patrik@ikus-soft.com>
"""
import os
from unittest import TestCase, mock, skipUnless

import cherrypy
import ldap3
from cherrypy.test import helper

from ..ldap import all_attribute, first_attribute  # noqa

original_connection = ldap3.Connection


def mock_ldap_connection(*args, **kwargs):
    kwargs.pop('client_strategy', None)
    return original_connection(*args, client_strategy=ldap3.MOCK_ASYNC, **kwargs)


class LdapFirstAttributeTest(TestCase):

    def test_no_keys_returns_default(self):
        attributes = {"cn": ["John Doe"]}
        self.assertIsNone(first_attribute(attributes, None))
        self.assertEqual(first_attribute(attributes, [], default="fallback"), "fallback")

    def test_single_key_with_scalar_value(self):
        attributes = {"uid": "jdoe"}
        self.assertEqual(first_attribute(attributes, "uid"), "jdoe")

    def test_single_key_with_list_value(self):
        attributes = {"mail": ["john@example.com", "alt@example.com"]}
        self.assertEqual(first_attribute(attributes, "mail"), "john@example.com")

    def test_empty_list_value_is_skipped(self):
        attributes = {
            "mail": [],
            "uid": ["jdoe"],
        }
        self.assertEqual(first_attribute(attributes, ["mail", "uid"]), "jdoe")

    def test_missing_first_key_uses_next_key(self):
        attributes = {"cn": ["John Doe"]}
        self.assertEqual(first_attribute(attributes, ["sn", "cn"]), "John Doe")

    def test_all_keys_missing_returns_default(self):
        attributes = {"cn": ["John Doe"]}
        self.assertEqual(first_attribute(attributes, ["sn", "uid"], default="unknown"), "unknown")

    def test_key_not_found_without_default_returns_none(self):
        attributes = {"cn": ["John Doe"]}
        self.assertIsNone(first_attribute(attributes, "uid"))


class LdapAllAttributeTest(TestCase):

    def test_no_keys_returns_default(self):
        attributes = {"cn": ["John Doe"]}
        self.assertIsNone(all_attribute(attributes, None))
        self.assertEqual(all_attribute(attributes, [], default="fallback"), "fallback")

    def test_single_key_with_scalar_value(self):
        attributes = {"uid": "jdoe"}
        self.assertEqual(all_attribute(attributes, "uid"), ["jdoe"])

    def test_single_key_with_list_value(self):
        attributes = {"mail": ["john@example.com", "alt@example.com"]}
        self.assertEqual(all_attribute(attributes, "mail"), ["john@example.com"])

    def test_multiple_keys_collect_values(self):
        attributes = {
            "uid": "jdoe",
            "cn": ["John Doe"],
        }
        self.assertEqual(all_attribute(attributes, ["uid", "cn"]), ["jdoe", "John Doe"])

    def test_missing_keys_are_skipped(self):
        attributes = {"cn": ["John Doe"]}
        self.assertEqual(all_attribute(attributes, ["sn", "cn", "uid"]), ["John Doe"])

    def test_all_keys_missing_returns_default(self):
        attributes = {"cn": ["John Doe"]}
        self.assertEqual(all_attribute(attributes, ["sn", "uid"], default=[]), [])

    def test_empty_list_value_is_appended_as_empty_list(self):
        attributes = {
            "mail": [],
            "uid": "jdoe",
        }
        self.assertEqual(all_attribute(attributes, ["mail", "uid"]), [[], "jdoe"])


class LdapPluginTest(helper.CPWebCase):

    @classmethod
    def setup_server(cls):
        # Configure Mock server early.
        cls.server = ldap3.Server('my_fake_server')
        cls.patcher = mock.patch('ldap3.Connection', side_effect=mock_ldap_connection)
        cls.patcher.start()
        cherrypy.config.update(
            {
                'ldap.uri': 'my_fake_server',
                'ldap.base_dn': 'dc=example,dc=org',
                'ldap.email_attribute': ['mail', 'email'],
                'ldap.fullname_attribute': ['displayName'],
                'ldap.firstname_attribute': ['givenName'],
                'ldap.lastname_attribute': ['sn'],
            }
        )

    @classmethod
    def teardown_class(cls):
        cherrypy.ldap.uri = None
        # Release mock server.
        cls.patcher.stop()
        return super().teardown_class()

    def setUp(self) -> None:
        # Clear LDAP entries before test.
        for entry in list(cherrypy.ldap._pool.strategy.connection.server.dit):
            if entry == 'cn=schema':
                continue
            cherrypy.ldap._pool.strategy.remove_entry(entry)
        return super().setUp()

    def test_authenticate(self):
        # Given a user in LDAP
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
            },
        )
        # When authenticating with that user
        authenticated = cherrypy.ldap.authenticate('user01', 'password1')
        # Then user is authenticated
        self.assertTrue(authenticated)
        self.assertEqual(authenticated[0], 'user01')
        self.assertEqual(
            {
                'dn': 'cn=user01,dc=example,dc=org',
                'userPassword': ['password1'],
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
                'cn': ['user01'],
                'email': None,
                'fullname': '',
            },
            authenticated[1],
        )

    def test_authenticate_with_wildcard(self):
        # Given a user in LDAP
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
            },
        )
        # When authenticating with a username containing wildcard
        authenticated = cherrypy.ldap.authenticate('user*', 'password1')
        # Then user is NOT authenticated
        self.assertFalse(authenticated)

    def test_authenticate_with_invalid_user(self):
        # Given a user in LDAP
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
            },
        )
        # When authenticating with an invalid user
        authenticated = cherrypy.ldap.authenticate('invalid', 'password1')
        # Then user is authenticated
        self.assertEqual(False, authenticated)

    def test_authenticate_with_invalid_password(self):
        # Given a user in LDAP
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
            },
        )
        # When authenticating with an invalid user
        authenticated = cherrypy.ldap.authenticate('user01', 'invalid')
        # Then user is not authenticated
        self.assertEqual(False, authenticated)
        # When authenticating with an valid password
        authenticated = cherrypy.ldap.authenticate('user01', 'password1')
        # Then user is not authenticated
        self.assertTrue(authenticated)

    def test_authenticate_with_email(self):
        # Given a user in LDAP with firstname and lastname
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'cn': ['user01'],
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
                'mail': ['john@test.com'],
            },
        )
        # When authenticating with that user
        authenticated = cherrypy.ldap.authenticate('user01', 'password1')
        # Then user is authenticated and firstname is composed of firstname and lastname
        self.assertTrue(authenticated)
        self.assertEqual(authenticated[0], 'user01')
        self.assertEqual(
            {
                'dn': 'cn=user01,dc=example,dc=org',
                'cn': ['user01'],
                'userPassword': ['password1'],
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
                'mail': ['john@test.com'],
                'email': 'john@test.com',
                'fullname': '',
            },
            authenticated[1],
        )

    def test_authenticate_with_fullname(self):
        # Given a user in LDAP with firstname and lastname
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'cn': ['user01'],
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
                'displayName': ['John Kennedy'],
            },
        )
        # When authenticating with that user
        authenticated = cherrypy.ldap.authenticate('user01', 'password1')
        # Then user is authenticated and firstname is composed of firstname and lastname
        self.assertTrue(authenticated)
        self.assertEqual(authenticated[0], 'user01')
        self.assertEqual(
            {
                'dn': 'cn=user01,dc=example,dc=org',
                'cn': ['user01'],
                'userPassword': ['password1'],
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
                'displayName': ['John Kennedy'],
                'email': None,
                'fullname': 'John Kennedy',
            },
            authenticated[1],
        )

    def test_authenticate_with_firstname_lastname(self):
        # Given a user in LDAP with firstname and lastname
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'cn': ['user01'],
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
                'givenName': ['John'],
                'sn': ['Kennedy'],
            },
        )
        # When authenticating with that user
        authenticated = cherrypy.ldap.authenticate('user01', 'password1')
        # Then user is authenticated and firstname is composed of firstname and lastname
        self.assertTrue(authenticated)
        self.assertEqual(authenticated[0], 'user01')
        self.assertEqual(
            {
                'dn': 'cn=user01,dc=example,dc=org',
                'cn': ['user01'],
                'userPassword': ['password1'],
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
                'givenName': ['John'],
                'sn': ['Kennedy'],
                'email': None,
                'fullname': 'John Kennedy',
            },
            authenticated[1],
        )


class LdapPluginTestWithRequiredGroup(helper.CPWebCase):

    @classmethod
    def setup_server(cls):
        # Configure Mock server early.
        cls.server = ldap3.Server('my_fake_server')
        cls.conn = ldap3.Connection(cls.server, client_strategy=ldap3.MOCK_ASYNC, raise_exceptions=True)
        cls.patcher = mock.patch('ldap3.Connection', return_value=cls.conn)
        cls.patcher.start()
        cherrypy.config.update(
            {
                'ldap.uri': 'my_fake_server',
                'ldap.base_dn': 'dc=example,dc=org',
                'ldap.required_group': ['appgroup', 'secondgroup'],
                'ldap.group_attribute': 'memberUid',
                'ldap.group_attribute_is_dn': False,
            }
        )

    @classmethod
    def teardown_class(cls):
        cherrypy.ldap.uri = None
        # Release mock server.
        cls.patcher.stop()
        return super().teardown_class()

    def setUp(self) -> None:
        # Clear LDAP entries before test.
        for entry in list(cherrypy.ldap._pool.strategy.connection.server.dit):
            if entry == 'cn=schema':
                continue
            cherrypy.ldap._pool.strategy.remove_entry(entry)
        return super().setUp()

    def test_authenticate_with_valid_group(self):
        # Given a user and a group in LDAP
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
            },
        )
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=appgroup,ou=Groups,dc=example,dc=org',
            {
                'cn': ['appgroup'],
                'memberUid': ['user01', 'user02'],
                'objectClass': ['posixGroup'],
            },
        )
        # When authenticating with that user
        authenticated = cherrypy.ldap.authenticate('user01', 'password1')
        # Then user is authenticated
        self.assertTrue(authenticated)
        self.assertEqual(authenticated[0], 'user01')
        self.assertEqual(
            {
                'dn': 'cn=user01,dc=example,dc=org',
                'userPassword': ['password1'],
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
                'cn': ['user01'],
                'email': None,
                'fullname': '',
            },
            authenticated[1],
        )

    def test_authenticate_with_not_member(self):
        # Given a user and a group in LDAP
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
            },
        )
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=appgroup,ou=Groups,dc=nodomain', {'memberUid': ['invalid', 'user02'], 'objectClass': ['posixGroup']}
        )
        # When authenticating with that user
        authenticated = cherrypy.ldap.authenticate('user01', 'password1')
        # Then user is not authenticated
        self.assertEqual(False, authenticated)

    def test_authenticate_with_invalid_group(self):
        # Given a user and a group in LDAP
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=user01,dc=example,dc=org',
            {
                'userPassword': 'password1',
                'uid': ['user01'],
                'objectClass': ['person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'],
            },
        )
        cherrypy.ldap._pool.strategy.add_entry(
            'cn=invalid,ou=Groups,dc=nodomain', {'memberUid': ['user01', 'user02'], 'objectClass': ['posixGroup']}
        )
        # When authenticating with that user
        authenticated = cherrypy.ldap.authenticate('user01', 'password1')
        # Then user is not authenticated
        self.assertEqual(False, authenticated)


@skipUnless(os.environ.get('TEST_LDAP_URI', None), "required TEST_LDAP_URI pointing to openldap server")
class LdapPluginTestWithOpenldap(helper.CPWebCase):
    """
    This test required a real openldap serer running with 'user01' and 'password1'.
    """

    @classmethod
    def setup_server(cls):
        cherrypy.config.update(
            {
                'ldap.uri': os.environ.get('TEST_LDAP_URI'),
                'ldap.base_dn': os.environ.get('TEST_LDAP_BASE_DN', 'dc=example,dc=org'),
            }
        )

    @classmethod
    def teardown_class(cls):
        cherrypy.ldap.uri = None
        return super().teardown_class()

    def test_authenticate(self):
        user = cherrypy.ldap.authenticate('user01', 'password1')
        self.assertTrue(user)
        self.assertEqual(user[0], 'user01')

    def test_authenticate_with_invalid_user(self):
        self.assertEqual(False, cherrypy.ldap.authenticate('josh', 'password'))



# Internationalisation tool for cherrypy
# Copyright (C) 2012-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Internationalization (i18n) and Localization (l10n) support for CherryPy.

This module provides a CherryPy tool that integrates GNU gettext and Babel
to handle language selection, translations, locale-aware formatting, and
timezone handling on a per-request basis.

The active language is resolved in the following order (highest priority first):

1. Language explicitly set with ``with i18n.preferred_lang():``
2. User-defined callback (``tools.i18n.func``)
3. HTTP ``Accept-Language`` request header
4. Default language configured via ``tools.i18n.default``

Translations are loaded using Babel and gettext, and the resolved locale is
available through ``i18n.get_translation()`` during request handling.

---------------------------------------------------------------------------
Basic usage
---------------------------------------------------------------------------

Within Python code, mark translatable strings using ``ugettext`` or
``ungettext``:

    from i18n import gettext as _, ngettext

    class MyController:
        @cherrypy.expose
        def index(self):
            locale = cherrypy.response.i18n.locale
            s1 = _(u"Translatable string")
            s2 = ngettext(
                u"There is one item.",
                u"There are multiple items.",
                2
            )
            return "<br />".join([s1, s2, locale.display_name])

---------------------------------------------------------------------------
Lazy translations
---------------------------------------------------------------------------

If code is executed before a CherryPy response object is available
(e.g. model definitions or module-level constants), use the ``*_lazy``
helpers. These defer translation until the value is actually rendered:

    from i18n_tool import gettext_lazy

    class Model:
        name = gettext_lazy(u"Model name")

---------------------------------------------------------------------------
Templates
---------------------------------------------------------------------------

For template rendering, i18n integrate with jinja2.

    {% trans %}Text to translate{% endtrans %}
    {{ _('Text to translate') }}
    {{ get_translation().gettext('Text to translate') }}
    {{ get_translation().locale }}
    {{ var | format_datetime(format='full') }}
    {{ var | format_date(format='full') }}

---------------------------------------------------------------------------
Configuration
---------------------------------------------------------------------------

Example CherryPy configuration:

    [/]
    tools.i18n.on = True
    tools.i18n.default = "en_US"
    tools.i18n.mo_dir = "/path/to/i18n"
    tools.i18n.domain = "myapp"

The ``mo_dir`` directory must contain subdirectories named after language
codes (e.g. ``en``, ``fr_CA``), each containing an ``LC_MESSAGES`` directory
with the compiled ``.mo`` file:

    <mo_dir>/<language>/LC_MESSAGES/<domain>.mo

"""

import logging
import os
from contextlib import contextmanager
from contextvars import ContextVar
from functools import lru_cache
from gettext import NullTranslations, translation

import cherrypy
import pytz
from babel import dates
from babel.core import Locale, get_global
from babel.support import LazyProxy, Translations

_preferred_lang = ContextVar('preferred_lang', default=())
_preferred_timezone = ContextVar('preferred_timezone', default=())
_translation = ContextVar('translation', default=None)
_tzinfo = ContextVar('tzinfo', default=None)


def _get_config(key, default=None):
    """
    Lookup configuration from request, if available. Fallback to global config.
    """
    if getattr(cherrypy, 'request') and getattr(cherrypy.request, 'config') and key in cherrypy.request.config:
        return cherrypy.request.config[key]
    return cherrypy.config.get(key, default)


@contextmanager
def preferred_lang(lang):
    """
    Re-define the preferred language to be used for translation within a given context.

    with i18n.preferred_lang('fr'):
        i18n.gettext('some string')
    """
    if not (lang is None or isinstance(lang, str)):
        raise ValueError(lang)
    try:
        # Update preferred lang and clear translation.
        if lang:
            token_l = _preferred_lang.set((lang,))
        else:
            token_l = _preferred_lang.set(tuple())
        token_t = _translation.set(None)
        yield
    finally:
        # Restore previous value
        _preferred_lang.reset(token_l)
        _translation.reset(token_t)


@contextmanager
def preferred_timezone(timezone):
    """
    Re-define the preferred timezone to be used for date format within a given context.

    with i18n.preferred_lang('America/Montreal'):
        i18n.format_datetime(...)
    """
    if not (timezone is None or isinstance(timezone, str)):
        raise ValueError(timezone)
    try:
        # Update preferred timezone and clear tzinfo.
        if timezone:
            token_t = _preferred_timezone.set((timezone,))
        else:
            token_t = _preferred_timezone.set(tuple())
        token_z = _tzinfo.set(None)
        yield
    finally:
        # Restore previous value
        _preferred_timezone.reset(token_t)
        _tzinfo.reset(token_z)


@lru_cache(maxsize=32)
def _search_translation(dirname, domain, *locales, sourcecode_lang='en'):
    """
    Loads the first existing translations for known locale.

    :parameters:
        langs : List
            List of languages as returned by `parse_accept_language_header`.
        dirname : String
            A single directory of the translations (`tools.i18n.mo_dir`).
        domain : String
            Gettext domain of the catalog (`tools.i18n.domain`).

    :returns: Translations, the corresponding Locale object.
    """
    if not isinstance(locales, (list, tuple)):
        locales = tuple(locales)

    # Loop on each locales to find the best matching translation.
    for locale in locales:
        try:
            # Use `gettext.translation()` instead of `gettext.find()` to chain translation fr_CA -> fr -> src.
            t = translation(domain=domain, localedir=dirname, languages=[locale], fallback=True, class_=Translations)
        except Exception:
            # If exception occur while loading the translation file. The file is probably corrupted.
            cherrypy.log(
                f'failed to load gettext catalog domain={domain} localedir={dirname} locale={locale}',
                context='I18N',
                severity=logging.WARNING,
                traceback=True,
            )
            continue
        if t.__class__ is NullTranslations and not locale.startswith(sourcecode_lang):
            # Continue searching if translation is not found.
            continue
        t.locale = Locale.parse(locale)
        return t
    # If translation file not found, return default
    t = NullTranslations()
    t.locale = Locale(sourcecode_lang)
    return t


def get_language_name(lang_code):
    """
    Translate the language code into it's language display name.
    """
    try:
        locale = Locale.parse(lang_code)
    except Exception:
        return lang_code
    translation = get_translation()
    return locale.get_language_name(translation.locale)


def get_timezone():
    """
    Get the best timezone information for the current context.

    The timezone returned is determined with the following priorities:

    * value of preferred_timezone()
    * tools.i18n.default_timezone
    * default server time.

    """
    # When tzinfo is defined, use it
    tzinfo = _tzinfo.get()
    if tzinfo is not None:
        return tzinfo
    # Otherwise search for a valid timezone.
    default = _get_config('tools.i18n.default_timezone')
    preferred_timezone = _preferred_timezone.get()
    if default and default not in preferred_timezone:
        preferred_timezone = (
            *preferred_timezone,
            default,
        )
    for timezone in preferred_timezone:
        try:
            tzinfo = dates.get_timezone(timezone)
            break
        except Exception:
            pass
    # If we can't find a valid timezone using the default and preferred value, fall back to server timezone.
    if tzinfo is None:
        tzinfo = dates.get_timezone(None)
    _tzinfo.set(tzinfo)
    return tzinfo


def get_translation():
    """
    Get the best translation for the current context.
    """
    # When translation is defined, use it
    translation = _translation.get()
    if translation is not None:
        return translation

    # Otherwise, we need to search the translation.
    # `preferred_lang` should always has a sane value within a cherrypy request because of hooks
    # But we also need to support calls outside cherrypy.
    sourcecode_lang = _get_config('tools.i18n.sourcecode_lang', 'en')
    default = _get_config('tools.i18n.default')
    preferred_lang = _preferred_lang.get()
    if default and default not in preferred_lang:
        preferred_lang = (
            *preferred_lang,
            default,
        )
    mo_dir = _get_config('tools.i18n.mo_dir')
    domain = _get_config('tools.i18n.domain', 'messages')
    translation = _search_translation(mo_dir, domain, *preferred_lang, sourcecode_lang=sourcecode_lang)
    _translation.set(translation)
    return translation


def list_available_locales():
    """
    Return a list of available translations.
    """
    return_value = []
    # Always return the source code locale.
    sourcecode_lang = _get_config('tools.i18n.sourcecode_lang', 'en')
    return_value.append(Locale.parse(sourcecode_lang))
    # Then scan language directory for more translation.
    mo_dir = _get_config('tools.i18n.mo_dir')
    domain = _get_config('tools.i18n.domain', 'messages')
    if not mo_dir:
        return
    for lang in os.listdir(mo_dir):
        if os.path.exists(os.path.join(mo_dir, lang, 'LC_MESSAGES', f'{domain}.mo')):
            try:
                return_value.append(Locale.parse(lang))
            except Exception:
                continue
    return return_value


def list_available_timezones():
    """
    Return list of available timezone.
    """
    # Babel only support a narrow list of timezone.
    babel_timezone = get_global('zone_territories').keys()
    return [t for t in pytz.all_timezones if t in babel_timezone]


# Public translation functions
def gettext(message):
    """Standard translation function. You can use it in all your exposed
    methods and everywhere where the response object is available.

    :parameters:
        message : Unicode
            The message to translate.

    :returns: The translated message.
    :rtype: Unicode
    """
    return get_translation().gettext(message)


ugettext = gettext


def ngettext(singular, plural, num):
    """Like ugettext, but considers plural forms.

    :parameters:
        singular : Unicode
            The message to translate in singular form.
        plural : Unicode
            The message to translate in plural form.
        num : Integer
            Number to apply the plural formula on. If num is 1 or no
            translation is found, singular is returned.

    :returns: The translated message as singular or plural.
    :rtype: Unicode
    """
    return get_translation().ngettext(singular, plural, num)


ungettext = ngettext


def gettext_lazy(message):
    """Like gettext, but lazy.

    :returns: A proxy for the translation object.
    :rtype: LazyProxy
    """

    def func():
        return get_translation().gettext(message)

    return LazyProxy(func, enable_cache=False)


def format_datetime(datetime=None, format='medium', tzinfo=None):
    """
    Wrapper around babel format_datetime to use current locale and current timezone.
    """
    return dates.format_datetime(
        datetime=datetime,
        format=format,
        locale=get_translation().locale,
        tzinfo=tzinfo or get_timezone(),
    )


def format_date(datetime=None, format='medium', tzinfo=None):
    """
    Wrapper around babel format_date to provide a default locale.
    """
    # To enforce the timezone and locale, make use of format_datetime for dates.
    return dates.format_datetime(
        datetime=datetime,
        format=dates.get_date_format(format),
        locale=get_translation().locale,
        tzinfo=tzinfo or get_timezone(),
    )


def get_timezone_name(tzinfo, width='long'):
    return dates.get_timezone_name(tzinfo, width=width, locale=get_translation().locale)


def _load_default(mo_dir, domain, default, **kwargs):
    """
    Initialize the language using the default value from the configuration.
    """
    # Clear current translation
    _preferred_lang.set(tuple())
    _preferred_timezone.set(tuple())
    # Clear current translation
    _translation.set(None)
    # Clear current timezone
    _tzinfo.set(None)


def _load_accept_language(**kwargs):
    """
    When running within a request, load the preferred language from Accept-Language header.
    """
    if cherrypy.request.headers.elements('Accept-Language'):
        # Sort language by quality
        languages = sorted(cherrypy.request.headers.elements('Accept-Language'), key=lambda x: x.qvalue, reverse=True)
        _preferred_lang.set(tuple(lang.value.replace('-', '_') for lang in languages))
        # Clear current translation
        _translation.set(None)


def _load_cookie_language(**kwargs):
    """
    Load preferred language from a request cookie.

    Expected cookie value formats:
    - en
    - en_US
    - en-US
    """
    # Skip this step if cookie name is not defined.
    cookie_name = _get_config('tools.i18n.cookie_name')
    if not cookie_name:
        return

    # Check if value defined in cookie.
    cookie = cherrypy.request.cookie
    if cookie_name not in cookie:
        return

    try:
        value = cookie[cookie_name].value.replace('-', '_')
    except Exception:
        return

    # Set preferred language and clear cached translation
    _preferred_lang.set((value,))
    _translation.set(None)


def _load_func_language(**kwargs):
    """
    When running a request where a current user is found, load preferred language from user preferences.
    """
    func = _get_config('tools.i18n.lang', default=_get_config('tools.i18n.func'))
    if not func:
        return
    try:
        lang = func()
    except Exception:
        return
    if not lang:
        return
    # Add custom lang to preferred_lang
    _preferred_lang.set((lang,))
    # Clear current translation
    _translation.set(None)


def _load_func_tzinfo(**kwargs):
    """
    When running a request, load the preferred timezone information from user preferences.
    """
    func = _get_config('tools.i18n.tzinfo')
    if not func:
        return
    try:
        tzinfo = func()
    except Exception:
        return
    if not tzinfo:
        return
    # Add custom lang to preferred_lang
    _preferred_timezone.set((tzinfo,))
    # Clear current translation
    _tzinfo.set(None)


def _set_content_language(**kwargs):
    """
    Sets the Content-Language response header (if not already set) to the
    language of `cherrypy.response.i18n.locale`.
    """
    if 'Content-Language' not in cherrypy.response.headers:
        # Only define the content language if the handler uses i18n module.
        translation = _translation.get()
        if translation:
            locale = translation.locale
            language_tag = f"{locale.language}-{locale.territory}" if locale.territory else locale.language
            cherrypy.response.headers['Content-Language'] = language_tag


class I18nTool(cherrypy.Tool):
    """Tool to integrate babel translations in CherryPy."""

    def __init__(self):
        super().__init__('before_handler', _load_default, 'i18n')

    def _setup(self):
        cherrypy.Tool._setup(self)
        # Attach additional hooks as different priority to update preferred lang with more accurate preferences.
        cherrypy.request.hooks.attach('before_handler', _load_accept_language, priority=60)
        cherrypy.request.hooks.attach('before_handler', _load_cookie_language, priority=70)
        cherrypy.request.hooks.attach('before_handler', _load_func_language, priority=75)
        cherrypy.request.hooks.attach('before_handler', _load_func_tzinfo, priority=75)
        cherrypy.request.hooks.attach('before_finalize', _set_content_language)


cherrypy.tools.i18n = I18nTool()



# SMTP Plugins for cherrypy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import email.utils
import re
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from xml.etree.ElementTree import fromstring, tostring

import cherrypy
from cherrypy.process.plugins import SimplePlugin


def _html2plaintext(html, encoding='utf-8'):
    """From an HTML text, convert the HTML to plain text.
    If @param body_id is provided then this is the tag where the
    body (not necessarily <body>) starts.
    """
    # (c) Fry-IT, www.fry-it.com, 2007
    # <peter@fry-it.com>
    # download here: http://www.peterbe.com/plog/html2plaintext
    assert isinstance(html, str)
    # &#160; are non-breaking space
    html = html.replace('&#160;', ' ')
    url_index = []
    try:
        tree = fromstring(html)
        tree = tree.find('body')
        i = 0
        for link in tree.findall('.//a'):
            url = link.get('href')
            if url:
                i += 1
                link.tag = 'span'
                link.text = '%s [%s]' % (link.text, i)
                url_index.append(url)
        html = tostring(tree, encoding=encoding).decode(encoding)
    except Exception:
        # Don't fail if the html is invalid.
        pass
    # \r char is converted into &#13;, must remove it
    html = html.replace('&#13;', '')
    # Remove new line & spaces defined for html formating.
    html = re.sub('\n *', '', html)
    # Replace tags
    html = html.replace('<strong>', '*').replace('</strong>', '*')
    html = html.replace('<b>', '*').replace('</b>', '*')
    html = html.replace('<h3>', '*').replace('</h3>', '*')
    html = html.replace('<h2>', '**').replace('</h2>', '**\n')
    html = html.replace('<h1>', '**').replace('</h1>', '**\n')
    html = html.replace('<em>', '/').replace('</em>', '/')
    html = html.replace('<tr>', '\n')
    html = html.replace('</p>', '\n')
    html = re.sub('<style[^>]*>[^<]*</style>', '', html)
    html = re.sub(r'<br\s*/?>', '\n', html)
    html = re.sub('<[^>]*>', '', html)
    html = re.sub(r'\n+', '\n', html)
    html = html.replace('&gt;', '>')
    html = html.replace('&lt;', '<')
    html = html.replace('&amp;', '&')

    # strip all lines
    html = '\n'.join([x.strip() for x in html.splitlines()])
    html = html.replace('\n' * 2, '\n')

    for i, url in enumerate(url_index):
        if i == 0:
            html += '\n\n'
        html += '[%s] %s\n' % (i + 1, url)

    return html.strip('\n')


def _formataddr(value):
    """
    Format the given value into a valid email address. Support raw string, tuple or a list of string.
    """
    if not value:
        return None
    if isinstance(value, str):
        return value
    if isinstance(value, (tuple, list)) and len(value) == 2 and isinstance(value[0], str) and isinstance(value[1], str):
        return email.utils.formataddr(value)
    if not isinstance(value, (tuple, list)):
        raise TypeError('expect a string, a tuple or a list of email address')
    return ', '.join(map(_formataddr, value))


def _send_message(server, encryption, username, password, msg):
    """
    Send message using SMTP.
    """
    host, unused, port = server.partition(':')
    if encryption == 'ssl':
        conn = smtplib.SMTP_SSL(host, port or 465)
    else:
        conn = smtplib.SMTP(host, port or 25)
    try:
        if encryption == 'starttls':
            conn.starttls()
        # Authenticate if required.
        if username:
            conn.login(username, password)
        conn.send_message(msg)
    finally:
        conn.quit()


class SmtpPlugin(SimplePlugin):

    server = None
    username = None
    password = None
    encryption = None
    email_from = None
    bcc = None

    def _create_msg(self, subject: str, message: str, to=None, cc=None, bcc=None, reply_to=None, headers={}):
        assert subject
        assert message
        assert to or bcc

        # Record the MIME types of both parts - text/plain and text/html.
        msg = MIMEMultipart('alternative')
        msg['Subject'] = str(subject)
        msg['From'] = _formataddr(self.email_from)
        if to:
            msg['To'] = _formataddr(to)
        if cc:
            msg['Cc'] = _formataddr(cc)
        bcc_list = []
        if self.bcc:
            bcc_list.append(_formataddr(self.bcc))
        if bcc:
            bcc_list.append(_formataddr(bcc))
        if bcc_list:
            msg['Bcc'] = ', '.join(bcc_list)
        if reply_to:
            msg['Reply-To'] = _formataddr(reply_to)
        msg['Message-ID'] = email.utils.make_msgid()
        if headers:
            for key, value in headers.items():
                msg[key] = value
        text = _html2plaintext(message)
        msg.attach(MIMEText(text, 'plain', 'utf8'))
        msg.attach(MIMEText(message, 'html', 'utf8'))
        return msg

    def start(self):
        self.bus.log('Start SMTP plugin')
        self.bus.subscribe("send_mail", self.send_mail)
        self.bus.subscribe("queue_mail", self.queue_mail)

    def stop(self):
        self.bus.log('Stop SMTP plugin')
        self.bus.unsubscribe("send_mail", self.send_mail)
        self.bus.unsubscribe("queue_mail", self.queue_mail)

    def graceful(self):
        """Reload of subscribers."""
        self.stop()
        self.start()

    def queue_mail(self, *args, **kwargs):
        """
        Queue mail to be sent.
        """
        # Skip sending email if smtp server is not configured.
        if not self.server:
            cherrypy.log('cannot send email because SMTP Server is not configured', context='SMTP')
            return
        if not self.email_from:
            cherrypy.log('cannot send email because SMTP From is not configured', context='SMTP')
            return
        msg = self._create_msg(*args, **kwargs)
        self.bus.publish(
            'scheduler:add_job_now',
            _send_message,
            server=self.server,
            encryption=self.encryption,
            username=self.username,
            password=self.password,
            msg=msg,
        )

    def send_mail(self, *args, **kwargs):
        """
        Reusable method to be called to send email to the user user.
        `user` user object where to send the email.
        """
        # Skip sending email if smtp server is not configured.
        if not self.server:
            cherrypy.log('cannot send email because SMTP Server is not configured', context='SMTP')
            return
        if not self.email_from:
            cherrypy.log('cannot send email because SMTP From is not configured', context='SMTP')
            return

        msg = self._create_msg(*args, **kwargs)
        _send_message(
            server=self.server, encryption=self.encryption, username=self.username, password=self.password, msg=msg
        )


# Register SMTP plugin
cherrypy.smtp = SmtpPlugin(cherrypy.engine)
cherrypy.smtp.subscribe()

cherrypy.config.namespaces['smtp'] = lambda key, value: setattr(cherrypy.smtp, key, value)



# Ratelimit tools for cherrypy
# Copyright (C) 2022-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import hashlib
import os
import tempfile
import threading
import time
from collections import namedtuple

import cherrypy

from cherrypy_foundation.sessions import session_lock

Tracker = namedtuple('Tracker', ['token', 'hits', 'timeout'])


class _DataStore:
    """
    Base class for rate limit data store
    """

    def __init__(self, **kwargs):
        self._locks = {}

    def get_and_increment(self, token, delay, hit=1):
        lock = self._locks.setdefault(token, threading.RLock())
        with lock:
            tracker = self._load(token)
            if tracker is None or tracker.timeout < time.time():
                tracker = Tracker(token=token, hits=0, timeout=int(time.time() + delay))
            tracker = tracker._replace(hits=tracker.hits + hit)
            self._save(tracker)
        return tracker.hits, tracker.timeout

    def _save(self, tracker):
        raise NotImplementedError

    def _load(self, token):
        raise NotImplementedError


class RamRateLimit(_DataStore):
    """
    Store rate limit information in memory.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._data = {}

    def _load(self, token):
        return self._data.get(token)

    def _save(self, tracker):
        self._data[tracker.token] = tracker

    def reset(self):
        self._data = {}


class FileRateLimit(_DataStore):
    PREFIX = 'ratelimit-'

    def __init__(self, storage_path, **kwargs):
        super().__init__(**kwargs)
        assert storage_path, 'FileRateLimit requires storage_path'
        self.storage_path = os.path.abspath(storage_path)
        os.makedirs(self.storage_path, exist_ok=True)

    def _path(self, token):
        # Hash the token to avoid max path limit and invalid chars.
        digest = hashlib.sha256(token.encode('utf-8')).hexdigest()
        return os.path.join(self.storage_path, self.PREFIX + digest + '.txt')

    def _load(self, token):
        path = self._path(token)
        try:
            with open(path, 'r', encoding='utf-8') as f:
                line = f.readline().strip()
        except FileNotFoundError:
            return None
        except OSError:
            return None

        if not line:
            return None

        parts = line.split()
        if len(parts) != 2:
            return None

        try:
            hits = int(parts[0])
            timeout = int(parts[1])
            if hits < 0 or timeout < 0:
                return None
        except ValueError:
            return None

        return Tracker(token=token, hits=hits, timeout=timeout)

    def _save(self, tracker):
        path = self._path(tracker.token)
        # Atomic write: temp file in same directory
        fd, tmp = tempfile.mkstemp(dir=self.storage_path)
        try:
            with os.fdopen(fd, 'w', encoding='utf-8') as f:
                f.write(f"{tracker.hits} {int(tracker.timeout)}\n")
                f.flush()
                os.fsync(f.fileno())
            os.replace(tmp, path)
            # On POSIX, chmod on first write
            try:
                os.chmod(path, 0o600)
            except OSError:
                pass
        finally:
            try:
                if os.path.exists(tmp):
                    os.remove(tmp)
            except Exception:
                pass

    def reset(self):
        for item in os.listdir(self.storage_path):
            if not item.startswith(self.PREFIX):
                continue
            fn = os.path.join(self.storage_path, item)
            if not os.path.isfile(fn):
                continue
            try:
                os.remove(fn)
            except OSError:
                pass


class Ratelimit(cherrypy.Tool):
    CONTEXT = 'TOOLS.RATELIMIT'

    _datastores = {}

    def __init__(self, priority=60):
        super().__init__('before_handler', self.check_ratelimit, 'ratelimit', priority)

    def _get_datastore(self, storage_class, **kwargs):
        """
        Create new datastore or return existing datastore.
        """
        # Default to RAM if storage class is not defined.
        if storage_class is None:
            kwargs = {}
            storage_class = RamRateLimit
        # Lookup for matching storage.
        key = (storage_class, str(kwargs))
        datastore = self._datastores.get(key)
        if datastore is None:
            # Create new storage if not found.
            self._datastores[key] = datastore = storage_class(**kwargs)
        return datastore

    def check_ratelimit(
        self,
        delay=3600,
        limit=25,
        return_status=429,
        logout=False,
        scope=None,
        methods=None,
        debug=False,
        hit=1,
        storage_class=None,
        **storage_kwargs,
    ):
        """
        Verify the ratelimit. By default return a 429 HTTP error code (Too Many Request). After 25 request within the same hour.

        Arguments:
            delay:         Time window for analysis in seconds
            limit:         Number of request allowed for an entry point
            return_status: HTTP Error code to return.
            logout:        True to logout user when limit is reached
            scope:         if specify, define the scope of rate limit. Default to path_info.
            methods:       if specify, only the methods in the list will be rate limited.
        """
        if delay <= 0:
            raise ValueError('Invalid delay: %s' % delay)
        if limit <= 0:
            return

        # Check if this 'method' should be rate limited
        request = cherrypy.request
        if methods is not None and request.method not in methods:
            if debug:
                cherrypy.log(f'skip rate limit for HTTP method {request.method}', context=self.CONTEXT)
            return

        # Create storage using storage class
        datastore = self._get_datastore(storage_class, **storage_kwargs)

        # Identifier: prefer authenticated user; else client IP
        identifier = getattr(cherrypy.serving.request, 'login', None) or cherrypy.request.remote.ip

        # Include method unless methods is explicitly a single method
        method_part = request.method
        if methods and len(methods) == 1:
            method_part = next(iter(methods))

        # Scope: allow explicit; else normalized path (e.g., strip trailing slash)
        path = request.path_info.rstrip('/') or '/'
        scope_value = scope or path

        token = f"{identifier}|{method_part}|{scope_value}"

        # Get hits count using datastore.
        hits, timeout = datastore.get_and_increment(token, delay, hit)
        if debug:
            cherrypy.log(f'check and increase limit token={token} limit={limit} hits={hits}', context=self.CONTEXT)

        # Verify user has not exceeded rate limit
        remaining = max(0, limit - hits)

        # Define headers
        cherrypy.response.headers['X-RateLimit-Limit'] = str(limit)
        cherrypy.response.headers['X-RateLimit-Remaining'] = str(remaining)
        cherrypy.response.headers['X-RateLimit-Reset'] = str(timeout)

        if limit < hits:  # block only after 'limit' successful requests
            cherrypy.log(f'block access to path_info={request.path_info}', context=self.CONTEXT)
            if logout:
                if hasattr(cherrypy.serving, 'session'):
                    with session_lock() as session:
                        session.clear()
                raise cherrypy.HTTPRedirect("/")
            raise cherrypy.HTTPError(return_status)

    def increase_hit(self, hit=1):
        """
        May be called directly by handlers to add a hit for the given request.
        """
        conf = cherrypy.tools.ratelimit._merged_args()
        conf['hit'] = hit
        cherrypy.tools.ratelimit.callable(**conf)

    def reset(self):
        """
        Used to reset the ratelimit. (for unit test)
        """
        for datastore in self._datastores.values():
            datastore.reset()


cherrypy.tools.ratelimit = Ratelimit()



# Cherrypy-foundation
# Copyright (C) 2020-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import logging

import cherrypy

HTML_ERROR_TEMPLATE = '''<!DOCTYPE html>
<html>
<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8"></meta>
    <title>%(status)s</title>
    <style type="text/css">
    #powered_by {
        margin-top: 20px;
        border-top: 2px solid black;
        font-style: italic;
    }

    #traceback {
        color: red;
    }
    </style>
</head>
    <body>
        <h2>%(status)s</h2>
        <p>%(message)s</p>
        <pre id="traceback">%(traceback)s</pre>
    <div id="powered_by">
      <span>
        Powered by <a href="http://www.cherrypy.dev">CherryPy %(version)s</a>
      </span>
    </div>
    </body>
</html>
'''


def error_page(status='', message='', traceback='', version=''):
    """
    Error page handler to handle Plain Text (text/plain), Json (application/json) or HTML (text/html) output.

    If available, uses Jina2 environment to generate error page using `error_page.html`.
    """
    # Log server error exception
    if status.startswith('500'):
        cherrypy.log(
            f'error page status={status} message={message}\n{traceback}', context='ERROR-PAGE', severity=logging.WARNING
        )

    # Replace message by generic one for 404. Default implementation leak path info.
    if status == '404 Not Found' and cherrypy.serving.request.path_info in message:
        message = 'Nothing matches the given URI'

    # Check expected response type.
    mtype = cherrypy.serving.response.headers.get('Content-Type') or cherrypy.tools.accept.callable(
        ['text/html', 'text/plain', 'application/json']
    )
    if mtype == 'text/plain':
        return message
    elif mtype == 'application/json':
        return json.dumps({'message': message, 'status': status})
    elif mtype == 'text/html':
        context = {'status': status, 'message': message, 'traceback': traceback, 'version': version}
        if hasattr(cherrypy.tools, 'jinja2'):
            # Try to build a nice error page with Jinja2 env
            try:
                return cherrypy.tools.jinja2.render_request(template='error_page.html', context=context)
            except Exception:
                cherrypy.log(
                    'fail to render error page with jinja2',
                    context='ERROR-PAGE',
                    severity=logging.ERROR,
                    traceback=True,
                )
        # Fallback to built-int HTML error page
        return HTML_ERROR_TEMPLATE % context

    # Fallback to raw error message.
    return message



# Cherrypy-foundation
# Copyright (C) 2020-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cherrypy
from markupsafe import Markup


class JinjaWidget:
    """
    Create field widget from Jinja2 templates.
    """

    filename = None

    def __init__(self, **options):
        self.options = options

    def __call__(self, field, **kwargs):
        env = cherrypy.request.config.get('tools.jinja2.env')
        kwargs = dict(self.options, **kwargs)
        # Support JinjaX
        if self.filename.endswith('.jinja'):
            catalog = env.globals['catalog']
            return catalog.irender(self.filename[0:-6], field=field, **kwargs)
        else:
            tmpl = env.get_template(self.filename)
            return Markup(tmpl.render(field=field, **kwargs))


class SwitchWidget(JinjaWidget):
    filename = 'SwitchWidget.jinja'


class SideBySideMultiSelect(JinjaWidget):
    filename = 'SideBySideMultiSelect.jinja'



# CherryPy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import gettext
import importlib.resources
import unittest
from datetime import datetime, timezone

import cherrypy
from cherrypy import _cpconfig
from cherrypy.test import helper

from .. import i18n

TEXT_EN = 'Some text to translate'
TEXT_FR = 'Du texte à traduire'


class TestI18n(unittest.TestCase):
    def setUp(self):
        self.mo_dir = importlib.resources.files(__package__) / 'locales'
        self.assertTrue(self.mo_dir.is_dir())
        cherrypy.request.config = _cpconfig.Config()

    def test_search_translation_en(self):
        # Load default translation return translation
        t = i18n._search_translation(self.mo_dir, 'messages', 'en')
        self.assertEqual("en", t.locale.language)
        # Test translation object
        self.assertEqual(TEXT_EN, t.gettext(TEXT_EN))

    def test_search_translation_fr(self):
        # Custom lang return translation.
        t = i18n._search_translation(self.mo_dir, 'messages', 'fr')
        self.assertIsInstance(t, gettext.GNUTranslations)
        self.assertEqual("fr", t.locale.language)
        # Test translation object
        self.assertEqual(TEXT_FR, t.gettext(TEXT_EN))

    def test_search_translation_invalid(self):
        # Load invalid translation return None
        t = i18n._search_translation(self.mo_dir, 'messages', 'tr')
        # Return Null translations.
        self.assertIn('NullTranslations', str(t.__class__))


class Root:

    @cherrypy.expose
    @cherrypy.tools.i18n(on=True)
    def index(self):
        return i18n.gettext(TEXT_EN)


class AbstractI18nTest(helper.CPWebCase):
    default_lang = None
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.config.update(
            {
                'tools.i18n.default': cls.default_lang,
                'tools.i18n.mo_dir': importlib.resources.files(__package__) / 'locales',
                'tools.i18n.domain': 'messages',
                'tools.i18n.cookie_name': 'locale',
            }
        )
        cherrypy.tree.mount(Root(), '/')


class TestI18nWebCase(AbstractI18nTest):

    def test_language_with_invalid(self):
        #  Query the page without login-in
        self.getPage("/", headers=[("Accept-Language", "invalid")])
        self.assertStatus('200 OK')
        self.assertHeaderItemValue("Content-Language", "en")
        self.assertInBody(TEXT_EN)

    def test_language_with_unknown(self):
        #  Query the page without login-in
        self.getPage("/", headers=[("Accept-Language", "it")])
        self.assertStatus('200 OK')
        self.assertHeaderItemValue("Content-Language", "en")
        self.assertInBody(TEXT_EN)

    def test_language_en(self):
        self.getPage("/", headers=[("Accept-Language", "en-US,en;q=0.8")])
        self.assertStatus('200 OK')
        self.assertHeaderItemValue("Content-Language", "en-US")
        self.assertInBody(TEXT_EN)

    def test_language_en_fr(self):
        self.getPage("/", headers=[("Accept-Language", "en-US,en;q=0.8,fr-CA;q=0.8")])
        self.assertStatus('200 OK')
        self.assertHeaderItemValue("Content-Language", "en-US")
        self.assertInBody(TEXT_EN)

    def test_language_fr(self):
        self.getPage("/")
        self.assertInBody(TEXT_EN)
        self.getPage("/", headers=[("Accept-Language", "fr-CA;q=0.8,fr;q=0.6")])
        self.assertStatus('200 OK')
        self.assertHeaderItemValue("Content-Language", "fr-CA")
        self.assertInBody(TEXT_FR)

    def test_language_en_US_POSIX(self):
        # When calling with locale variant
        self.getPage("/", headers=[("Accept-Language", "en-US-POSIX")])
        self.assertStatus('200 OK')
        # Tehn page return en-US
        self.assertHeaderItemValue("Content-Language", "en-US")

    def test_cookie_fr(self):
        # When calling with locale variant
        self.getPage("/", headers=[("Accept-Language", "en-US"), ("Cookie", "locale=fr")])
        self.assertStatus('200 OK')
        # Tehn page return en-US
        self.assertHeaderItemValue("Content-Language", "fr")

    def test_with_preferred_lang(self):
        # Given a default lang 'en'
        date = datetime.fromtimestamp(1680111611, timezone.utc)
        self.assertEqual(TEXT_EN, i18n.ugettext(TEXT_EN))
        self.assertIn('March', i18n.format_datetime(date, format='long'))
        # When using preferred_lang with french
        with i18n.preferred_lang('fr'):
            # Then french translation is used
            self.assertEqual(TEXT_FR, i18n.ugettext(TEXT_EN))
            # Then date time formating used french locale
            self.assertIn('mars', i18n.format_datetime(date, format='long'))
        # Then ouside the block, settings goest back to english
        self.assertEqual(TEXT_EN, i18n.ugettext(TEXT_EN))
        self.assertIn('March', i18n.format_datetime(date, format='long'))

    def test_format_datetime_locales(self):
        date = datetime.fromtimestamp(1680111611, timezone.utc)
        with i18n.preferred_timezone('utc'):
            with i18n.preferred_lang('fr'):
                self.assertEqual('29 mars 2023, 17:40:11 TU', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('en'):
                self.assertEqual('March 29, 2023, 5:40:11\u202fPM UTC', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('en_US'):
                self.assertEqual('March 29, 2023, 5:40:11\u202fPM UTC', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('en_GB'):
                self.assertEqual('29 March 2023, 17:40:11 UTC', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('en_CH'):
                self.assertEqual('29 March 2023, 17:40:11 UTC', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('de'):
                self.assertEqual('29. März 2023, 17:40:11 UTC', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('de_CH'):
                self.assertEqual('29. März 2023, 17:40:11 UTC', i18n.format_datetime(date, format='long'))

        with i18n.preferred_timezone('Europe/Paris'):
            with i18n.preferred_lang('fr'):
                self.assertEqual('29 mars 2023, 19:40:11 +0200', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('fr_CH'):
                self.assertEqual('29 mars 2023, 19:40:11 +0200', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('en'):
                self.assertEqual('March 29, 2023, 7:40:11\u202fPM +0200', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('en_US'):
                self.assertEqual('March 29, 2023, 7:40:11\u202fPM +0200', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('en_GB'):
                self.assertEqual('29 March 2023, 19:40:11 CEST', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('en_CH'):
                self.assertEqual('29 March 2023, 19:40:11 CEST', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('de'):
                self.assertEqual('29. März 2023, 19:40:11 MESZ', i18n.format_datetime(date, format='long'))
            with i18n.preferred_lang('de_CH'):
                self.assertEqual('29. März 2023, 19:40:11 MESZ', i18n.format_datetime(date, format='long'))

    def test_list_available_locales(self):
        self.assertEqual(['de', 'en', 'fr'], sorted([str(l) for l in i18n.list_available_locales()]))

    def test_list_available_timezones(self):
        timezones = i18n.list_available_timezones()
        self.assertIn('America/Toronto', timezones)

    def test_get_timezone_name(self):
        with i18n.preferred_lang('en'):
            self.assertEqual('Eastern Time', i18n.get_timezone_name('America/Toronto'))
            self.assertEqual('ET', i18n.get_timezone_name('America/Toronto', width='short'))
            self.assertEqual('Eastern Time', i18n.get_timezone_name('America/Toronto', width='long'))
        with i18n.preferred_lang('fr'):
            self.assertEqual('heure de l’Est nord-américain', i18n.get_timezone_name('America/Toronto'))
            self.assertEqual('HE', i18n.get_timezone_name('America/Toronto', width='short'))
            self.assertEqual('heure de l’Est nord-américain', i18n.get_timezone_name('America/Toronto', width='long'))


class TestI18nDefaultLangWebCase(AbstractI18nTest):
    default_lang = 'FR'

    @classmethod
    def teardown_class(cls):
        # Reset default-lang to avoid issue with other test
        cherrypy.config['tools.i18n.default'] = 'en'
        super().teardown_class()

    def test_default_lang_without_accept_language(self):
        # Given a default language
        # When user connect to the application without Accept-Language
        self.getPage("/")
        self.assertStatus(200)
        # Then page is displayed with default lang
        self.assertInBody(TEXT_FR)

    def test_default_lang_with_accept_language(self):
        # Given a default language
        # When user connect to the application with Accept-Language English
        self.getPage("/", headers=[("Accept-Language", "en-US,en;q=0.8")])
        self.assertStatus(200)
        # Then page is displayed as english
        self.assertInBody(TEXT_EN)

    def test_default_lang_with_unknown_accept_language(self):
        # Given a default language
        # When user connect to the application with Accept-Language English
        self.getPage("/", headers=[("Accept-Language", "it")])
        self.assertStatus(200)
        # Then page is displayed as english
        self.assertInBody(TEXT_FR)


class TestI18nInvalidDefaultLangWebCase(AbstractI18nTest):
    default_lang = 'invalid'

    def test_default_lang_invalid(self):
        # Given an invalid default language
        # When user connect to the application without Accept-Language
        self.getPage("/")
        self.assertStatus(200)
        # Then page is displayed with fallback to "en"
        self.assertInBody(TEXT_EN)



# Cherrypy-foundation
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import tempfile
import unittest
from contextlib import contextmanager

from selenium import webdriver


class SeleniumUnitTest:

    @property
    def _session_id(self):
        if hasattr(self, 'cookies') and self.cookies:
            for unused, value in self.cookies:
                for part in value.split(';'):
                    key, unused, value = part.partition('=')
                    if key == 'session_id':
                        return value

    @contextmanager
    def selenium(self, headless=True, implicitly_wait=3):
        """
        Decorator to load selenium for a test.
        """
        # Skip selenium test is display is not available.
        if not os.environ.get('DISPLAY', False):
            raise unittest.SkipTest("selenium require a display")
        # Start selenium driver
        options = webdriver.ChromeOptions()
        if headless:
            options.add_argument('--headless')
        options.add_argument('--disable-gpu')
        options.add_argument('--window-size=1280,800')
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-dev-shm-usage')
        options.add_argument('--lang=en-US')
        driver = webdriver.Chrome(options=options)
        try:
            # If logged in, reuse the same session id.
            if self._session_id:
                driver.get(f'{self.baseurl}/login/')
                driver.add_cookie({"name": "session_id", "value": self.session_id})
            # Configure download folder
            download = os.path.join(os.path.expanduser('~'), 'Downloads')
            os.makedirs(download, exist_ok=True)
            self._selenium_download_dir = tempfile.mkdtemp(dir=download, prefix='selenium-download-')
            driver.execute_cdp_cmd(
                'Page.setDownloadBehavior', {'behavior': 'allow', 'downloadPath': self._selenium_download_dir}
            )
            # Set default wait.
            driver.implicitly_wait(implicitly_wait)
            yield driver
        finally:
            # Code to release resource, e.g.:
            driver.close()
            driver = None



# RestAPI plugin for cherrypy
# Copyright (C) 2024-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cherrypy


class Dispatcher(cherrypy.dispatch.Dispatcher):
    """
    Dispatcher using HTTP method to find the proper function to be called.

    e.g.:
    GET /api/users            -> list()
    GET /api/users/3          -> get(3)
    GET /api/users/my/path    -> get('my/path')
    DELETE /api/users/3       -> delete(3)
    DELETE /api/users/my/path -> delete('my/path')

    POST /api/users     -> post(data)
    POST /api/users/3   -> post(3, data)
    PUT /api/users      -> post(data)
    PUT /api/users/3    -> post(3, data)

    """

    supported_method = ['get', 'delete', 'post', 'put']

    def __call__(self, path):
        request = cherrypy.serving.request
        # To simplify implementation reuse default dispatcher function.
        # But strip the last segment
        resource, vpath = self.find_handler(path)

        if not resource:
            request.handler = cherrypy.NotFound()
            return

        # Two scenario possible. Either we have found our GET handler.
        # Or we need to look into the resource for it.
        meth = request.method.lower()
        if meth not in self.supported_method:
            request.handler = cherrypy.HTTPError(405)
        if meth == 'get' and not hasattr(resource, 'list') and not hasattr(resource, 'get'):
            request.handler = cherrypy.dispatch.LateParamPageHandler(resource, *vpath)
            return

        # Call "list()" instead of "get()" when path doesn't have an id or name.
        if meth == 'get' and not vpath and hasattr(resource, 'list'):
            meth = 'list'
        # Find the subhandler
        func = getattr(resource, meth, None)
        if func:
            # Grab any _cp_config on the subhandler.
            if hasattr(func, '_cp_config'):
                request.config.update(func._cp_config)

            # Decode any leftover %2F in the virtual_path atoms.
            if vpath:
                vpath = ['/'.join([x.replace('%2F', '/') for x in vpath])]
            request.handler = cherrypy.dispatch.LateParamPageHandler(func, *vpath)
        else:
            request.handler = cherrypy.HTTPError(405)



# Cherrypy-foundation
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import hashlib
import os
from base64 import b64decode, b64encode

try:
    import argon2

    _argon = argon2.PasswordHasher()

    def hash_password(password):
        assert password and isinstance(password, str)
        return _argon.hash(password)

except ImportError:
    _argon = None

    def hash_password(password):
        assert password and isinstance(password, str)
        password = password.encode(encoding='utf8')
        salt = os.urandom(4)
        h = hashlib.sha1(password)
        h.update(salt)
        return "{SSHA}" + b64encode(h.digest() + salt).decode('latin1')


def check_password(password, challenge):
    """
    Check if the password matches the challenge.
    The challenge is an encrypted password.
    """
    if not password or not challenge:
        return False
    assert isinstance(password, str)
    assert isinstance(challenge, str)
    if _argon and challenge.startswith('$argon2'):
        try:
            return _argon.verify(challenge, password)
        except Exception:
            return False
    elif challenge.startswith('{SSHA}'):
        digest_salt = b64decode(challenge[6:])
        digest = digest_salt[:20]
        sha = hashlib.sha1(password.encode(encoding='utf8'))
        sha.update(digest_salt[20:])
        return digest == sha.digest()
    else:
        # Fallback to previous SHA
        sha = hashlib.sha1(password.encode('utf8'))
        return challenge == sha.hexdigest()



# LDAP Plugins for cherrypy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import cherrypy
import ldap3
from cherrypy.process.plugins import SimplePlugin
from ldap3.core.exceptions import LDAPException, LDAPInvalidCredentialsResult

_safe = ldap3.utils.conv.escape_filter_chars


def all_attribute(attributes, keys, default=None):
    """
    Extract all values from LDAP attributes.
    """
    # Skip loopkup if key is not defined.
    if not keys:
        return default
    # Convert key to a list.
    keys = keys if isinstance(keys, list) else [keys]
    # Loop on each attribute name.
    values = []
    for attr in keys:
        try:
            value = attributes[attr]
            if isinstance(value, list) and len(value) > 0:
                values.append(value[0])
            else:
                values.append(value)
        except KeyError:
            pass
    # Default to None.
    return values if values else default


def first_attribute(attributes, keys, default=None):
    """
    Extract the first value from LDAP attributes.
    """
    # Skip loopkup if key is not defined.
    if not keys:
        return default
    # Convert key to a list.
    keys = keys if isinstance(keys, list) else [keys]
    # Loop on each attribute name to find a value.
    for attr in keys:
        try:
            value = attributes[attr]
            if isinstance(value, list):
                if len(value) == 0:
                    continue
                return value[0]
            else:
                return value
        except KeyError:
            pass
    # Default to None.
    return default


class LdapPlugin(SimplePlugin):
    """
    Used this plugin to authenticate user against an LDAP server.

    `authenticate(username, password)` return None if the credentials are not
    valid. Otherwise it return a tuple or username and extra attributes.
    The extra attribute may contains `_fullname` and `_email`.
    """

    uri = None
    base_dn = None
    bind_dn = ''
    bind_password = ''
    scope = 'subtree'
    tls = False
    user_filter = '(objectClass=*)'
    username_attribute = ['uid']
    required_group = None
    group_attribute = 'member'
    group_attribute_is_dn = False
    group_filter = '(objectClass=*)'
    version = 3
    network_timeout = 10
    timeout = 10
    fullname_attribute = None
    firstname_attribute = None
    lastname_attribute = None
    email_attribute = None
    pool_size = 10

    def start(self):
        # Don't configure this plugin if the ldap URI is not provided.
        if not self.uri:
            return
        self.bus.log('Start LDAP connection')
        # Set up the LDAP server object
        server = ldap3.Server(self.uri, connect_timeout=self.network_timeout, mode=ldap3.IP_V4_PREFERRED)
        # Create a pool of reusable connections
        self._pool = ldap3.Connection(
            server,
            user=self.bind_dn,
            password=self.bind_password,
            auto_bind=ldap3.AUTO_BIND_TLS_BEFORE_BIND if self.tls else ldap3.AUTO_BIND_NO_TLS,
            version=self.version,
            raise_exceptions=True,
            client_strategy=ldap3.REUSABLE,
            pool_size=self.pool_size,
        )

    def stop(self):
        self.bus.log('Stop LDAP connection')
        # Release the connection pool.
        if hasattr(self, '_pool'):
            self._pool.unbind()

    def graceful(self):
        """Reload of subscribers."""
        self.stop()
        self.start()

    def authenticate(self, username, password):
        """
        Check if the given credential as valid according to LDAP.
        Return False if invalid.
        Return None if the plugin is unavailable to validate credentials or if the plugin is disabled.
        Return tuple (<username>, <attributes>) if the credentials are valid.
        """
        assert isinstance(username, str)
        assert isinstance(password, str)

        if not hasattr(self, '_pool'):
            return None

        # Connect to LDAP Server.
        with self._pool as conn:
            try:
                # Search the LDAP server for user's DN.
                safe_username = _safe(username)
                attr_filter = ''.join([f'({_safe(attr)}={safe_username})' for attr in self.username_attribute])
                search_filter = f"(&{self.user_filter}(|{attr_filter}))"
                response = self._search(conn, search_filter)
                if not response:
                    cherrypy.log(f"lookup failed username={username} reason=not_found", context='LDAP')
                    return False
                cherrypy.log(f"lookup successful username={username}", context='LDAP')
                user_dn = response[0]['dn']

                # Use a separate connection to validate credentials
                login_conn = ldap3.Connection(
                    self._pool.server,
                    user=user_dn,
                    password=password,
                    version=self.version,
                    raise_exceptions=True,
                    client_strategy=ldap3.ASYNC,
                )
                if not login_conn.bind():
                    cherrypy.log(
                        f'ldap authentication failed username={username} reason=wrong_password',
                        context='LDAP',
                        severity=logging.WARNING,
                    )
                    return False

                # Get username
                attrs = response[0]['attributes']
                new_username = first_attribute(attrs, self.username_attribute)
                if not new_username:
                    cherrypy.log(
                        f"object missing username attribute user_dn={user_dn} attribute={self.username_attribute}",
                        context='LDAP',
                        severity=logging.WARNING,
                    )
                    return False

                # Verify if the user is member of the required group
                if self.required_group:
                    if not isinstance(self.required_group, list):
                        self.required_group = [self.required_group]
                    user_value = user_dn if self.group_attribute_is_dn else new_username
                    group_filter = '(&(%s=%s)(|%s)%s)' % (
                        _safe(self.group_attribute),
                        _safe(user_value),
                        ''.join(['(cn=%s)' % _safe(group) for group in self.required_group]),
                        self.group_filter,
                    )
                    # Search LDAP Server for matching groups.
                    cherrypy.log(
                        f"group check start username={user_value} required_groups={' '.join(self.required_group)}",
                        context='LDAP',
                    )
                    response = self._search(conn, group_filter, attributes=['cn'])
                    if not response:
                        cherrypy.log(
                            f"group check failed username={user_value} required_groups={' '.join(self.required_group)}",
                            context='LDAP',
                        )
                        return False

                # Extract common attribute from LDAP
                attrs['dn'] = user_dn
                attrs['email'] = first_attribute(attrs, self.email_attribute)
                attrs['fullname'] = fullname = first_attribute(attrs, self.fullname_attribute)
                if not fullname:
                    firstname = first_attribute(attrs, self.firstname_attribute, '')
                    lastname = first_attribute(attrs, self.lastname_attribute, '')
                    attrs['fullname'] = ' '.join([name for name in [firstname, lastname] if name])
                return (new_username, attrs)
            except LDAPInvalidCredentialsResult:
                return False
            except LDAPException:
                cherrypy.log(
                    f"unexpected error username={username}", context='LDAP', severity=logging.ERROR, traceback=True
                )
            return None

    def search(self, filter, attributes=ldap3.ALL_ATTRIBUTES, search_base=None, paged_size=None):
        with self._pool as conn:
            return self._search(
                conn, filter=filter, attributes=attributes, search_base=search_base, paged_size=paged_size
            )

    def _search(self, conn, filter, attributes=ldap3.ALL_ATTRIBUTES, search_base=None, paged_size=None):
        search_scope = {'base': ldap3.BASE, 'onelevel': ldap3.LEVEL, 'subtree': ldap3.SUBTREE}.get(
            self.scope, ldap3.SUBTREE
        )
        search_base = search_base or self.base_dn
        cherrypy.log(f"search {self.uri}/{search_base}?{search_scope}?{filter}", context='LDAP')
        msg_id = conn.search(
            search_base=search_base,
            search_filter=filter,
            search_scope=search_scope,
            time_limit=self.timeout,
            attributes=attributes,
            paged_size=paged_size,
        )
        response, _result_unused = conn.get_response(msg_id)
        return response


cherrypy.ldap = LdapPlugin(cherrypy.engine)
cherrypy.ldap.subscribe()

cherrypy.config.namespaces['ldap'] = lambda key, value: setattr(cherrypy.ldap, key, value)



# Secure Headers tool for cherrypy
# Copyright (C) 2012-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import http.cookies

import cherrypy

DEFAULT_CSP = {
    'default-src': 'self',
    'style-src': ('self', 'unsafe-inline'),
    'script-src': ('self', 'unsafe-inline'),
}

#
# Patch Morsel prior to 3.8
# Allow SameSite attribute to be define on the cookie.
#
if not http.cookies.Morsel().isReservedKey("samesite"):
    http.cookies.Morsel._reserved['samesite'] = 'SameSite'


def _build_csp(csp):
    if isinstance(csp, dict):
        return "; ".join(
            f"{directive} {' '.join(sources) if isinstance(sources, (list, tuple)) else str(sources)}"
            for directive, sources in csp.items()
        )
    return str(csp)


def set_headers(
    xfo='DENY',
    no_cache=True,
    referrer='same-origin',
    nosniff=True,
    xxp='1; mode=block',
    csp=DEFAULT_CSP,
):
    """
    This tool provide CSRF mitigation.

    * Define X-Frame-Options = DENY
    * Define Cookies SameSite=Lax
    * Define Cookies Secure when https is detected
    * Validate `Origin` and `Referer` on POST, PUT, PATCH, DELETE
    * Define Cache-Control by default
    * Define Referrer-Policy to 'same-origin'

    Ref.:
    https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html
    https://cheatsheetseries.owasp.org/cheatsheets/Clickjacking_Defense_Cheat_Sheet.html
    """
    request = cherrypy.request
    response = cherrypy.serving.response

    # Check if Origin matches our target.
    if request.method in ['POST', 'PUT', 'PATCH', 'DELETE']:
        origin = request.headers.get('Origin', None)
        if origin and origin != request.base:
            raise cherrypy.HTTPError(403, 'Unexpected Origin header')

    # Check if https is enabled
    https = request.base.startswith('https')

    # Define X-Frame-Options to avoid Clickjacking
    if xfo:
        response.headers['X-Frame-Options'] = xfo

    # Enforce security on cookies
    cookie = response.cookie.get('session_id', None)
    if cookie:
        # Awaiting bug fix in cherrypy
        # https://github.com/cherrypy/cherrypy/issues/1767
        # Force SameSite to Lax
        cookie['samesite'] = 'Lax'
        if https:
            cookie['secure'] = 1

    # Add Cache-Control to avoid storing sensible information in Browser cache.
    if no_cache:
        response.headers['Cache-control'] = 'no-cache, no-store, must-revalidate, max-age=0'
        response.headers['Pragma'] = 'no-cache'
        response.headers['Expires'] = '0'

    # Add Referrer-Policy
    if referrer:
        response.headers['Referrer-Policy'] = referrer

    # Add X-Content-Type-Options to avoid browser to "sniff" to content-type
    if nosniff:
        response.headers['X-Content-Type-Options'] = 'nosniff'

    # Add X-XSS-Protection to enabled XSS protection
    if xxp:
        response.headers['X-XSS-Protection'] = xxp

    # Add Content-Security-Policy
    if csp:
        response.headers['Content-Security-Policy'] = _build_csp(csp)

    # Add Strict-Transport-Security to force https use.
    if https:
        response.headers['Strict-Transport-Security'] = "max-age=31536000; includeSubDomains"


cherrypy.tools.secure_headers = cherrypy.Tool('before_request_body', set_headers, priority=71)



# Cherrypy-foundation
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import logging.handlers
import sys

import cherrypy


def _coerce_level(level):
    if isinstance(level, int):
        return level
    if isinstance(level, str):
        lvl = logging._nameToLevel.get(level.upper())
        if isinstance(lvl, int):
            return lvl
    raise ValueError(f"Invalid log level: {level!r}")


def _remove_cherrypy_date(record):
    """Remove the leading date for cherrypy error."""
    if record.name.startswith('cherrypy.error'):
        record.msg = record.msg[23:].strip()
    return True


def _add_ip(record):
    """Add request IP to record."""
    # Check if we are serving a real request
    if cherrypy.request and cherrypy.request.request_line:
        remote = cherrypy.request.remote
        record.ip = remote.name or remote.ip
    else:
        record.ip = "-"
    return True


def _add_username(record):
    """Add current username to record."""
    # Check if we are serving a real request
    if cherrypy.request and cherrypy.request.request_line:
        if cherrypy.request.login:
            record.user = cherrypy.request.login
        else:
            record.user = "anonymous"
    else:
        record.user = "-"
    return True


def setup_logging(log_file, log_access_file, level):
    """
    Configure the logging system for CherryPy.
    Safe to call multiple times.
    """
    lvl = _coerce_level(level)

    cherrypy.config.update({'log.screen': False, 'log.access_file': '', 'log.error_file': ''})
    cherrypy.engine.unsubscribe('graceful', cherrypy.log.reopen_files)

    # Configure root logger
    logger = logging.getLogger('')
    logger.level = lvl
    # Capture warnings
    logging.captureWarnings(True)
    if log_file:
        print("continue logging to %s" % log_file)
        default_handler = logging.handlers.RotatingFileHandler(log_file, maxBytes=10485760, backupCount=20)
    else:
        default_handler = logging.StreamHandler(sys.stdout)
    default_handler.addFilter(_remove_cherrypy_date)
    default_handler.addFilter(_add_ip)
    default_handler.addFilter(_add_username)
    default_handler.setFormatter(
        logging.Formatter("[%(asctime)s][%(levelname)-7s][%(ip)s][%(user)s][%(threadName)s][%(name)s] %(message)s")
    )
    logger.addHandler(default_handler)

    # Configure cherrypy access logger
    cherrypy_access = logging.getLogger('cherrypy.access')
    cherrypy_access.propagate = False
    if log_access_file:
        handler = logging.handlers.RotatingFileHandler(log_access_file, maxBytes=10485760, backupCount=20)
        cherrypy_access.addHandler(handler)

    # Configure cherrypy error logger
    cherrypy_error = logging.getLogger('cherrypy.error')
    cherrypy_error.propagate = False
    cherrypy_error.addHandler(default_handler)



# CherryPy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import tempfile

import cherrypy
from cherrypy.test import helper

from .. import ratelimit  # noqa


class Root:

    @cherrypy.expose
    def index(self):
        return "OK"

    @cherrypy.expose
    @cherrypy.tools.ratelimit(on=True, methods=['POST'])
    def login(self):
        return "login"


class RateLimitTest(helper.CPWebCase):
    interactive = False
    rate_limit = 5
    rate_limit_dir = None

    def setUp(self):
        cherrypy.tools.ratelimit.reset()
        return super().setUp()

    def tearDown(self):
        cherrypy.tools.ratelimit.reset()
        return super().tearDown()

    @classmethod
    def setup_server(cls):
        rate_limit_storage_class = None
        if cls.rate_limit_dir:
            rate_limit_storage_class = ratelimit.FileRateLimit
        cherrypy.config.update(
            {
                'tools.ratelimit.debug': True,
                'tools.ratelimit.delay': 3600,
                'tools.ratelimit.limit': cls.rate_limit,
                'tools.ratelimit.storage_class': rate_limit_storage_class,
                'tools.ratelimit.storage_path': cls.rate_limit_dir,
            }
        )
        cherrypy.tree.mount(Root(), '/')

    def test_ratelimit(self):
        # Given a endpoint with ratelimit enabled
        # When requesting multiple time the page
        for i in range(0, 5):
            self.getPage('/login', method='POST')
            self.assertStatus(200)
        # Then a 429 error (too many request) is return
        self.getPage('/login', method='POST')
        self.assertStatus(429)

    def test_ratelimit_forwarded_for(self):
        # Given a endpoint with ratelimit enabled
        # When requesting multiple time the login page with different `X-Forwarded-For`
        for i in range(0, 5):
            self.getPage(
                '/login',
                headers=[('X-Forwarded-For', '127.0.0.%s' % i)],
                method='POST',
            )
            self.assertStatus(200)
        # Then original IP get blocked
        self.getPage(
            '/login',
            headers=[('X-Forwarded-For', '127.0.0.%s' % i)],
            method='POST',
        )
        self.assertHeaderItemValue('X-Ratelimit-Limit', '5')
        self.assertHeaderItemValue('X-Ratelimit-Remaining', '0')
        self.assertHeader('X-Ratelimit-Reset')
        self.assertStatus(429)


class FileStorageRateLimitTest(RateLimitTest):

    @classmethod
    def setup_class(cls):
        cls.tempdir = tempfile.TemporaryDirectory(prefix='cherrypy-foundation-', suffix='-ratelimit-test')
        cls.rate_limit_dir = cls.tempdir.name
        super().setup_class()

    @classmethod
    def teardown_class(cls):
        cls.tempdir.cleanup()
        super().teardown_class()



# CherryPy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
from collections import namedtuple
from urllib.parse import urlencode

import cherrypy
from cherrypy.lib.sessions import RamSession
from cherrypy.test import helper

from cherrypy_foundation.sessions import session_lock

from ..auth import AUTH_LAST_PASSWORD_AT
from ..auth_mfa import (
    MFA_CODE_TIME,
    MFA_DEFAULT_CODE_TIMEOUT,
    MFA_DEFAULT_TRUST_DURATION,
    MFA_TRUSTED_IP_LIST,
    MFA_USER_KEY,
    MFA_VERIFICATION_TIME,
)
from ..sessions_timeout import SESSION_PERSISTENT

User = namedtuple('User', 'id,username,password,email,mfa', defaults=[False])

users = {
    User(2, 'myuser', 'changeme', 'myuser@example.com', False),
    User(3, 'mfauser', 'changeme', 'mfauser@example.com', True),
    User(4, 'noemail', 'changeme', '', True),
}


def checkpassword(username, password):
    for u in users:
        if u.username == username and u.password == password:
            return True
    return False


def user_lookup_func(login, user_info):
    for u in users:
        if u.username == login:
            return u.id, u
    return None


def user_from_key_func(userkey):
    for u in users:
        if u.id == userkey:
            return u
    return None


@cherrypy.tools.sessions(locking='explicit')
@cherrypy.tools.auth(
    user_lookup_func=user_lookup_func,
    user_from_key_func=user_from_key_func,
    checkpassword=checkpassword,
)
@cherrypy.tools.auth_mfa(
    mfa_enabled=lambda: hasattr(cherrypy.serving.request, 'currentuser') and cherrypy.request.currentuser.mfa
)
class Root:

    @cherrypy.expose
    def index(self):
        return "OK"

    @cherrypy.expose()
    def login(self, username=None, password=None):
        if cherrypy.serving.request.method == 'POST' and username and password:
            userobj = cherrypy.tools.auth.login_with_credentials(username, password)
            if userobj:
                raise cherrypy.tools.auth.redirect_to_original_url()
            else:
                return "invalid credentials"
        return "login"

    @cherrypy.expose()
    def mfa(self, code=None, resend_code=None, persistent=False):
        html = "mfa\n"
        if cherrypy.serving.request.method == 'POST' and code:
            if cherrypy.tools.auth_mfa.verify_code(code=code, persistent=persistent):
                raise cherrypy.tools.auth.redirect_to_original_url()
            else:
                html += "<p>invalid verification code</p>\n"
        # Send verification code if previous code expired.
        # Or when requested by user.
        if (resend_code and cherrypy.serving.request.method == 'POST') or cherrypy.tools.auth_mfa.is_code_expired():
            code = cherrypy.tools.auth_mfa.generate_code()
            # Here the code should be send by email, SMS or any other means.
            # For our test we store it in the session.
            with session_lock() as session:
                session['code'] = code
            html += "<p>new verification code sent to your email</p>\n"
        return html


class AuthManagerMfaTest(helper.CPWebCase):
    interactive = False
    # Authenticated by default.
    login = True

    @classmethod
    def setup_server(cls):
        cherrypy.tree.mount(Root(), '/')

    def getPage(self, *args, **kwargs):
        """
        This implementation keep track of session cookies.
        """
        headers = kwargs.pop('headers', [])
        if hasattr(self, 'cookies') and self.cookies:
            headers.extend(self.cookies)
        return helper.CPWebCase.getPage(self, *args, headers=headers, **kwargs)

    def _login(self, username, password):
        self.getPage('/login/', method='POST', body=urlencode({'username': username, 'password': password}))
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/')

    @property
    def _session_id(self):
        if hasattr(self, 'cookies') and self.cookies:
            for unused, value in self.cookies:
                for part in value.split(';'):
                    key, unused, value = part.partition('=')
                    if key == 'session_id':
                        return value

    def _get_code(self):
        # Query MFA page to generate a code
        self.getPage("/mfa/")
        self.assertStatus(200)
        self.assertInBody("new verification code sent to your email")
        # Extract code from user session for testing only.
        session = RamSession.cache[self._session_id][0]
        return session['code']

    def test_get_without_login(self):
        # Given the user is not authenticated.
        # When requesting /mfa/
        self.getPage("/mfa/")
        # Then user is redirected to /login/
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/login/')

    def test_get_with_mfa_disabled(self):
        # Given an authenticated user with MFA Disable
        self._login('myuser', 'changeme')
        # When requesting /mfa/ page
        self.getPage("/mfa/")
        # Then user is redirected to root page
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/')
        # Then index is enabled.
        self.getPage("/")
        self.assertStatus(200)
        self.assertInBody('OK')

    def test_get_with_trusted(self):
        # Given an authenticated user with MFA Disable
        self._login('mfauser', 'changeme')
        # Given an authenticated user with MFA enabled and already verified
        session = RamSession.cache[self._session_id][0]
        session[MFA_USER_KEY] = 3
        session[MFA_VERIFICATION_TIME] = datetime.datetime.now()
        session[MFA_TRUSTED_IP_LIST] = ['127.0.0.1']

        # When requesting /mfa/ page when we are already trusted
        self.getPage("/mfa/")
        # Then user is redirected to root page
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/')

    def test_get_with_trusted_expired(self):
        # Given an authenticated user with MFA enabled and already verified
        self._login('mfauser', 'changeme')
        session = RamSession.cache[self._session_id][0]
        session[MFA_USER_KEY] = 3
        session[MFA_VERIFICATION_TIME] = datetime.datetime.now() - datetime.timedelta(minutes=60)

        # When requesting /mfa/ page
        self.getPage("/mfa/")
        self.assertStatus(200)
        # Then a verification code is send to the user
        self.assertInBody("new verification code sent to your email")

    def test_get_with_trusted_different_ip(self):
        # Given an authenticated user with MFA enabled and already verified
        self._login('mfauser', 'changeme')
        session = RamSession.cache[self._session_id][0]
        session[MFA_USER_KEY] = 3
        session[MFA_VERIFICATION_TIME] = datetime.datetime.now()

        # When requesting /mfa/ page from a different ip
        self.getPage("/mfa/", headers=[('X-Forwarded-For', '10.255.14.23')])
        self.assertStatus(200)
        # Then a verification code is send to the user
        self.assertInBody("new verification code sent to your email")

    def test_get_without_verified(self):
        # Given an authenticated user With MFA enabled
        self._login('mfauser', 'changeme')
        # When requesting /mfa/ page
        self.getPage("/mfa/")
        self.assertStatus(200)
        # Then a verification code is send to the user
        self.assertInBody("new verification code sent to your email")

    def test_verify_code_valid(self):
        prev_session_id = self._session_id
        # Given an authenticated user With MFA enabled
        self._login('mfauser', 'changeme')
        code = self._get_code()
        # When sending a valid verification code
        self.getPage("/mfa/", method='POST', body=urlencode({'code': code}))
        # Then a new session_id is generated
        self.assertNotEqual(prev_session_id, self._session_id)
        # Then user is redirected to root page
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/')
        # Then user has access
        self.getPage("/")
        self.assertStatus(200)

    def test_verify_code_invalid(self):
        # Given an authenticated user With MFA enabled
        # When sending an invalid verification code
        self.getPage("/mfa/", method='POST', body=urlencode({'code': '1234567'}))
        # Then user is redirected to login page
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/login/')

    def test_verify_code_expired(self):
        # Given an authenticated user With MFA enabled
        self._login('mfauser', 'changeme')
        code = self._get_code()
        # When sending a valid verification code that expired
        session = RamSession.cache[self._session_id][0]

        session[MFA_CODE_TIME] = datetime.datetime.now() - datetime.timedelta(minutes=MFA_DEFAULT_CODE_TIMEOUT + 1)

        self.getPage("/mfa/", method='POST', body=urlencode({'code': code}))
        # Then a new code get generated.
        self.assertStatus(200)
        self.assertInBody("invalid verification code")

    def test_verify_code_invalid_after_3_tentative(self):
        # Given an authenticated user With MFA
        self._login('mfauser', 'changeme')
        code = self._get_code()
        # When user enter an invalid verification code 3 times
        self.getPage("/mfa/", method='POST', body=urlencode({'code': '1234567'}))
        self.assertStatus(200)
        self.getPage("/mfa/", method='POST', body=urlencode({'code': '1234567'}))
        self.assertStatus(200)
        self.getPage("/mfa/", method='POST', body=urlencode({'code': '1234567'}))
        # Then an error get displayed to the user
        self.assertStatus(200)
        self.assertInBody("invalid verification code")
        # Then a new code get send to the user.
        self.assertInBody("new verification code sent to your email")
        session = RamSession.cache[self._session_id][0]
        new_code = session['code']
        self.assertNotEqual(code, new_code)

    def test_resend_code(self):
        # Given an authenticated user With MFA enabled with an existing code
        self._login('mfauser', 'changeme')
        code = self._get_code()
        # When user request a new code
        self.getPage("/mfa/", method='POST', body=urlencode({'resend_code': '1'}))
        # Then A success message is displayedto the user.
        self.assertInBody("new verification code sent to your email")
        session = RamSession.cache[self._session_id][0]
        new_code = session['code']
        self.assertNotEqual(code, new_code)

    def test_redirect_to_original_url(self):
        # Given an authenticated user
        self._login('mfauser', 'changeme')
        # When querying a page that required mfa
        self.getPage('/prefs/general')
        # Then user is redirected to mfa page
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/mfa/')
        # When providing verification code
        code = self._get_code()
        self.getPage("/mfa/", method='POST', body=urlencode({'code': code}))
        # Then user is redirected to original url
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/prefs/general')

    def test_login_persistent_when_login_timeout(self):
        prev_session_id = self._session_id
        # Given a user authenticated with MFA with "persistent"
        self._login('mfauser', 'changeme')
        code = self._get_code()
        self.getPage("/mfa/", method='POST', body=urlencode({'code': code, 'persistent': '1'}))
        self.assertStatus(303)
        self.getPage("/")
        self.assertStatus(200)
        self.assertNotEqual(prev_session_id, self._session_id)
        session = RamSession.cache[self._session_id][0]
        self.assertTrue(session[SESSION_PERSISTENT])
        # When the re-auth time expired (after 15 min)
        session[AUTH_LAST_PASSWORD_AT] = datetime.datetime.now() - datetime.timedelta(minutes=60, seconds=1)

        # Then next query redirect user to /login/ page (by mfa)
        self.getPage("/")
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/login/')
        prev_session_id = self._session_id
        # When user enter valid username password
        self.getPage("/login/", method='POST', body=urlencode({'username': 'mfauser', 'password': 'changeme'}))
        # Then user is redirected to original url without need to pass MFA again.
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/')
        self.assertNotEqual(prev_session_id, self._session_id)
        self.getPage("/")
        self.assertStatus(200)
        self.assertInBody('OK')

    def test_login_persistent_when_mfa_timeout(self):
        prev_session_id = self._session_id
        # Given a user authenticated with MFA with "persistent"
        self._login('mfauser', 'changeme')
        code = self._get_code()
        self.getPage("/mfa/", method='POST', body=urlencode({'code': code, 'persistent': '1'}))
        self.assertStatus(303)
        self.getPage("/")
        self.assertStatus(200)
        self.assertNotEqual(prev_session_id, self._session_id)
        session = RamSession.cache[self._session_id][0]

        self.assertTrue(session[SESSION_PERSISTENT])
        # When the mfa verification timeout (after 15 min)
        session[MFA_VERIFICATION_TIME] = datetime.datetime.now() - datetime.timedelta(
            minutes=MFA_DEFAULT_TRUST_DURATION, seconds=1
        )

        # Then next query redirect user to mfa page
        self.getPage("/prefs/general")
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/mfa/')
        # When user enter valid code
        code = self._get_code()
        self.getPage("/mfa/", method='POST', body=urlencode({'code': code, 'persistent': '1'}))
        # Then user is redirected to original page.
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/prefs/general')
        self.getPage("/")
        self.assertStatus(200)
        self.assertInBody('OK')



# Authentication tools for cherrypy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import logging
import urllib.parse

import cherrypy

from cherrypy_foundation.sessions import session_lock

AUTH_LAST_PASSWORD_AT = '_auth_last_password_at'
AUTH_METHOD = '_auth_method'
AUTH_ORIGINAL_URL = '_auth_original_url'

AUTH_DEFAULT_REDIRECT = "/login/"
AUTH_DEFAULT_SESSION_KEY = "_auth_session_key"
AUTH_DEFAULT_REAUTH_TIMEOUT = 60  # minutes


class AuthManager(cherrypy.Tool):
    """
    CherryPy tool handling authentication.

    Required config:
      - tools.auth.user_lookup_func(login, user_info) -> (user_key, userobj)
      - tools.auth.user_from_key_func(user_key) -> userobj | None
      - tools.auth.checkpassword: callable or list of callables (login, password) ->
          False | (login, user_info) | str(login)
    """

    def __init__(self):
        super().__init__(point='before_handler', callable=self._restore_from_session, priority=70)

    def _setup(self):
        cherrypy.Tool._setup(self)
        # Attach additional hooks as different priority to update preferred lang with more accurate preferences.
        conf = self._merged_args()
        conf.pop('priority', None)
        cherrypy.serving.request.hooks.attach('before_handler', self._forbidden_or_redirect, priority=74, **conf)

    # ---- Config helpers ----

    @staticmethod
    def _reauth_timeout_minutes() -> int:
        # How long until we force a full username/password re-login regardless of session timeout.
        return int(cherrypy.request.config.get('tools.auth.reauth_timeout', AUTH_DEFAULT_REAUTH_TIMEOUT))

    @staticmethod
    def _session_user_key() -> int:
        return cherrypy.request.config.get('tools.auth.session_user_key', AUTH_DEFAULT_SESSION_KEY)

    # ---- Tool entrypoints ----

    def _restore_from_session(self, **kwargs):
        """
        Early hook: attempt to restore 'authenticated' state from session.
        """
        # Verify if session is enabled, if not the user is not authenticated.
        if not hasattr(cherrypy.serving, 'session'):
            return
        # Check if a user_key is stored in session.
        with session_lock() as session:
            user_key = session.get(self._session_user_key())
            if not user_key:
                return

            last = session.get(AUTH_LAST_PASSWORD_AT)
            if last is None:
                return  # never had a password login in this session
            timeout = self._reauth_timeout_minutes()
            if (last + datetime.timedelta(minutes=timeout)) < session.now():
                return

        # Mark request as authenticated by key; user object will be resolved later.
        cherrypy.serving.request.login = user_key

    def _forbidden_or_redirect(self, user_from_key_func, redirect=AUTH_DEFAULT_REDIRECT, **kwargs):
        """
        If authenticated via session, resolve user object; else redirect/403.
        """
        # Allow access to the login page itself
        if cherrypy.serving.request.path_info == redirect:
            return

        user_key = getattr(cherrypy.serving.request, 'login', False)
        if user_key:
            try:
                currentuser = user_from_key_func(user_key)
            except Exception:
                cherrypy.log(
                    f'unexpected error searching for user_key={user_key}',
                    context='AUTH',
                    severity=logging.ERROR,
                    traceback=True,
                )
                currentuser = None

            if currentuser:
                cherrypy.serving.request.currentuser = currentuser
                return

        # Not authenticated or user not found.
        if redirect:
            self.save_original_url()
            raise cherrypy.HTTPRedirect(redirect)
        raise cherrypy.HTTPError(403)

    # ---- Login flows ----

    def login_with_credentials(self, login, password):
        """
        Validate credentials with configured checkers; on success, call login_with_result.
        """
        if not login or not password:
            cherrypy.log('authentication failed reason=empty_credentials', context='AUTH', severity=logging.WARNING)
            return None
        # Validate credentials using checkpassword function(s).
        conf = self._merged_args()
        checkpassword = conf.get('checkpassword')
        if not isinstance(checkpassword, (list, tuple)):
            checkpassword = [checkpassword]
        for func in checkpassword:
            try:
                valid = func(login, password)
                if not valid:
                    continue
                # Support various return value: tuple, string with username, boolean value
                if isinstance(valid, (list, tuple)) and len(valid) >= 2:
                    login, user_info = valid
                elif isinstance(valid, str):
                    login, user_info = valid, None
                else:
                    login, user_info = login, None
                # If authentication is successful, initiate login process.
                return self.login_with_result(login=login, user_info=user_info)
            except Exception:
                cherrypy.log(
                    f'unexpected error checking password login={login} checkpassword={func.__qualname__} - continue with next function',
                    context='AUTH',
                    severity=logging.ERROR,
                    traceback=True,
                )
        # If we reach here, authentication failed
        if hasattr(cherrypy.serving, 'session'):
            with session_lock() as session:
                session.regenerate()  # Prevent session analysis

        remote_ip = cherrypy.serving.request.remote.ip
        cherrypy.log(
            f'authentication failed login={login} ip={remote_ip} reason=wrong_credentials',
            context='AUTH',
            severity=logging.WARNING,
        )

        return None

    def login_with_result(self, login=None, user_info=None, auth_method='password'):
        """
        Called after credentials were validated or via SSO.
        Resolves user via user_lookup_func and establishes the session.
        """
        conf = self._merged_args()
        user_lookup_func = conf.get('user_lookup_func')

        try:
            user_key, userobj = user_lookup_func(login=login, user_info=user_info or {})
        except Exception:
            cherrypy.log(
                f"unexpected error searching user login={login} user_info={user_info}",
                context='AUTH',
                severity=logging.ERROR,
                traceback=True,
            )
            return None

        if not userobj:
            cherrypy.log(
                f"authentication failed login={login} reason=not_found", context='AUTH', severity=logging.WARNING
            )
            return None

        # Notify plugins about user login
        cherrypy.engine.publish('user_login', userobj)

        # Store in session
        if hasattr(cherrypy.serving, 'session'):
            with session_lock() as session:
                session_user_key = self._session_user_key()
                session[session_user_key] = user_key
                session[AUTH_METHOD] = auth_method
                session[AUTH_LAST_PASSWORD_AT] = session.now()
                # Generate a new session id
                session.regenerate()

        # When authenticated, store user_key in request.
        cherrypy.serving.request.login = user_key
        return userobj

    # ---- Session helpers ----

    def clear_session(self):
        """
        Clear session data and generate a new session id.
        """
        with session_lock() as session:
            session.clear()
            session.regenerate()

    def get_original_url(self):
        """
        Return the original URL browsed by the user before authentication.
        """
        if not hasattr(cherrypy.serving, 'session'):
            return None
        with session_lock() as session:
            return session.get(AUTH_ORIGINAL_URL)

    def get_user_key(self):
        """Return the last username."""
        if not hasattr(cherrypy.serving, 'session'):
            return False
        session_user_key = self._session_user_key()
        with session_lock() as session:
            return session.get(session_user_key)

    def redirect_to_original_url(self):
        # Redirect user to original URL
        redirect_url = self.get_original_url() or '/'
        return cherrypy.HTTPRedirect(redirect_url)

    def save_original_url(self):
        """
        Save the current URL to user's session.
        """
        # Skip this step if session is not enabled
        if not hasattr(cherrypy.serving, 'session'):
            return
        # Extract URL including query-string.
        request = cherrypy.serving.request
        uri_encoding = getattr(request, 'uri_encoding', 'utf-8')
        original_url = urllib.parse.quote(request.path_info, encoding=uri_encoding)
        query_string = request.query_string

        # Store value in session
        with session_lock() as session:
            session[AUTH_ORIGINAL_URL] = cherrypy.url(original_url, qs=query_string, base='')


cherrypy.tools.auth = AuthManager()



# CherryPy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import tempfile
from pathlib import Path

import cherrypy
from cherrypy.test import helper

from ..logging import setup_logging


class Root:

    @cherrypy.expose
    def index(self):
        return "OK"

    @cherrypy.expose
    def error(self):
        cherrypy.log('error messages to be logged')


class LoggingTest(helper.CPWebCase):
    interactive = False

    @classmethod
    def setup_server(cls):
        cls.tempdir = tempfile.TemporaryDirectory(prefix='cherrypy-foundation-', suffix='-logging')
        cls.log_access_file = f'{cls.tempdir.name}/access.log'
        cls.log_file = f'{cls.tempdir.name}/error.log'
        setup_logging(log_file=cls.log_file, log_access_file=cls.log_access_file, level='DEBUG')
        cherrypy.tree.mount(Root(), '/')

    @classmethod
    def teardown_class(cls):
        # Delete temp folder
        cls.tempdir.cleanup()
        # Stop server
        super().teardown_class()
        # Reset logging to default.
        # Re-enable screen logging
        cherrypy.config.update({'log.screen': True, 'log.error_file': '', 'log.access_file': ''})
        # Reset internal logger references
        cherrypy.log.error_file = None
        cherrypy.log.access_file = None

    def test_logging_access(self):
        mtime = Path(self.log_access_file).stat().st_mtime
        # When page get queried
        self.getPage('/')
        # Then access files get updated
        mtime2 = Path(self.log_access_file).stat().st_mtime
        self.assertNotEqual(mtime, mtime2)

    def test_logging_error(self):
        data = Path(self.log_file).read_text()
        self.assertNotIn('error messages to be logged', data)
        # When page get queried
        self.getPage('/error')
        self.assertStatus(200)
        # Then access files get updated
        data2 = Path(self.log_file).read_text()
        self.assertNotEqual(data, data2)
        self.assertIn('error messages to be logged', data2)



# Scheduler plugins for Cherrypy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import copy
import logging
from datetime import datetime, timezone
from functools import wraps
from threading import Event, RLock

import cherrypy
from apscheduler.events import EVENT_JOB_ERROR, EVENT_JOB_EXECUTED, EVENT_JOB_SUBMITTED
from apscheduler.schedulers.background import BackgroundScheduler
from cherrypy.process.plugins import SimplePlugin

logger = logging.getLogger(__name__)


def clear_db_sessions(func):
    """
    A decorator that ensures database connections that have become unusable, or are obsolete, are closed
    before and after the job is executed.
    """

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        finally:
            cherrypy.db.clear_sessions()
        return result

    return func_wrapper


class Scheduler(SimplePlugin):
    """
    Plugins to run Job at fixed time (cronjob) and to schedule task to be run on by one.

    Configuration:
    - jobstore. Take value 'memory' (default), 'db' to use SQLalchemy database in combination with cherrypy.db. Or any callable function to create an instance of BaseJobStore
    """

    jobstores = {"default": {"type": "memory"}}
    executors = {"default": {"type": "threadpool", "max_workers": 10}}
    job_defaults = {"coalesce": False, "max_instances": 1}
    timezone = None

    def __init__(self, bus):
        super().__init__(bus)
        self._scheduler = None
        # Let keep track of running jobs manually.
        # This is required for unit testing to wait for all running task.
        self._lock = RLock()
        self._event = Event()
        self._submitted_jobs = set()
        self._completed_jobs = set()

    @property
    def _running_jobs(self):
        with self._lock:
            return self._submitted_jobs

    def _track_running_jobs(self, event):
        """
        This listener keep track of running jobs.
        This is required to implement wait for tasks mostly used during unit testing.
        """
        with self._lock:
            # Special care must be taken to handle race condition
            # the EVENT_JOB_EXECUTED may get called before EVENT_JOB_SUBMITTED
            if event.code == EVENT_JOB_SUBMITTED:
                if event.job_id in self._completed_jobs:
                    self._completed_jobs.remove(event.job_id)
                else:
                    self._submitted_jobs.add(event.job_id)
            elif event.code in (EVENT_JOB_EXECUTED, EVENT_JOB_ERROR):
                if event.job_id in self._submitted_jobs:
                    self._submitted_jobs.remove(event.job_id)
                else:
                    self._completed_jobs.add(event.job_id)
            # Wakeup wait_for_tasks.
            self._event.set()

    def _drop_table(self, target, conn, **kwargs):
        """
        Callback listener to drop table.
        This is required because SQLAlchemyJobStore creates the table manually.
        """
        from sqlalchemy.sql import ddl

        # Here we need to use private API to loop over all sqlalchemy job store.
        for jobstore in self._scheduler._jobstores.values():
            table = getattr(jobstore, 'jobs_t', None)
            if table is not None:
                conn.execute(ddl.DropTable(table, if_exists=True))

    def _create_table(self, target, conn, **kwargs):
        """
        Callback listener to drop table.
        This is required because SQLAlchemyJobStore creates the table manually.
        """
        from sqlalchemy.sql import ddl

        # Here we need to use private API to loop over all sqlalchemy job store.
        for jobstore in self._scheduler._jobstores.values():
            table = getattr(jobstore, 'jobs_t', None)
            if table is not None:
                conn.execute(ddl.CreateTable(table, if_not_exists=True))

    def _db_events(self, listen_or_remove):
        """
        Listen on database events to create and drop tables as required.
        """
        assert listen_or_remove in ['listen', 'remove']
        try:
            from sqlalchemy import event

            if getattr(cherrypy, 'db', False):
                base = cherrypy.db.get_base()
                func = getattr(event, listen_or_remove)
                func(base.metadata, 'after_drop', self._drop_table)
                func(base.metadata, 'after_create', self._create_table)
        except ImportError:
            # Nothing to do if sqlalchemy is not available
            pass

    def _create_scheduler(self):
        """
        Create a new APScheduler base on plugin configuration.
        """
        # Create the scheduler with deepcopy of configuration.
        scheduler = BackgroundScheduler(
            jobstores=copy.deepcopy(self.jobstores),
            executors=copy.deepcopy(self.executors),
            job_defaults=copy.deepcopy(self.job_defaults),
            timezone=self.timezone,
        )
        # Add listener to keep track of running jobs.
        scheduler.add_listener(self._track_running_jobs, (EVENT_JOB_EXECUTED | EVENT_JOB_ERROR | EVENT_JOB_SUBMITTED))
        # Start the scheduler.
        scheduler.start()
        return scheduler

    def start(self):
        self.bus.log('Start Scheduler plugin')
        # Create apscheduler
        self._scheduler = self._create_scheduler()
        # Register drop_table event (when using database)
        self._db_events('listen')
        # Then register new channels
        self.bus.subscribe('scheduler:add_job', self.add_job)
        self.bus.subscribe('scheduler:add_job_daily', self.add_job_daily)
        self.bus.subscribe('scheduler:add_job_now', self.add_job_now)
        self.bus.subscribe('scheduler:remove_job', self.remove_job)

    # Slightly lower priority to start after database, but before other plugins.
    start.priority = 49

    def stop(self):
        self.bus.log('Stop Scheduler plugin')
        # Unregister drop_table event (when using database)
        self._db_events('remove')
        # Shutdown the scheduler & wait for running task to complete.
        if self._scheduler:
            self._scheduler.shutdown(wait=True)
            self._scheduler = None
        self.bus.unsubscribe('scheduler:add_job', self.add_job)
        self.bus.unsubscribe('scheduler:add_job_daily', self.add_job_daily)
        self.bus.unsubscribe('scheduler:add_job_now', self.add_job_now)
        self.bus.unsubscribe('scheduler:remove_job', self.remove_job)

    stop.priority = 51

    def graceful(self):
        """Reload of subscribers."""
        self.stop()
        self.start()

    def add_job(self, func, *args, **kwargs):
        """
        Called via engine.publish('scheduler:add_job', ...)
        OR directly on the plugin instance.
        """
        if not self._scheduler:
            raise RuntimeError("Scheduler not running; cannot add job.")

        return self._scheduler.add_job(func, *args, **kwargs)

    def add_job_daily(self, execution_time, func, *args, **kwargs):
        """
        A convenience wrapper around add_job() that ensures
        the job runs daily.
        """
        hour, minute = execution_time.split(':', 2)
        return self.add_job(
            id=getattr(func, '__name__', str(func)),
            func=func,
            name=getattr(func, '__name__', str(func)),
            args=args,
            kwargs=kwargs,
            trigger='cron',
            hour=hour,
            minute=minute,
            misfire_grace_time=None,
            coalesce=True,
            replace_existing=True,
        )

    def add_job_now(self, func, *args, **kwargs):
        """
        A convenience wrapper around add_job() that ensures
        the job runs immediately upon being added.
        """
        return self.add_job(
            func=func,
            name=getattr(func, '__name__', str(func)),
            args=args,
            kwargs=kwargs,
            next_run_time=datetime.now(timezone.utc),
            misfire_grace_time=None,
        )

    def get_jobs(self, jobstore=None):
        """
        Return list of scheduled jobs.
        """
        if not self._scheduler:
            raise RuntimeError("Scheduler not running; cannot get jobs.")
        return self._scheduler.get_jobs(jobstore=jobstore)

    def remove_job(self, job, jobstore=None):
        """
        Remove the given job from scheduler.
        """
        # Search for a matching job
        return_value = False
        if self._scheduler is None:
            return return_value
        for j in self._scheduler.get_jobs(jobstore=jobstore):
            if j.func == job or j.name == job:
                self._scheduler.remove_job(job_id=j.id, jobstore=jobstore)
                return_value = True
        return return_value

    def remove_all_jobs(self, jobstore=None):
        """
        Remove all jobs from scheduler.
        """
        if self._scheduler is None:
            return
        self._scheduler.remove_all_jobs(jobstore=jobstore)

    def wait_for_jobs(self, jobstore=None):
        """
        Used to wait for all running jobs to complete.
        """
        if self._scheduler is None:
            return
        # Wait until the queue is empty.
        while any(
            job for job in self._scheduler.get_jobs(jobstore=jobstore) if job.next_run_time < datetime.now(timezone.utc)
        ):
            self._event.wait(timeout=1)
            self._event.clear()
        # Wait until all jobs are completed.
        while self._running_jobs:
            self._event.wait(timeout=1)
            self._event.clear()


# Register Scheduler plugin
cherrypy.scheduler = Scheduler(cherrypy.engine)
cherrypy.scheduler.subscribe()

cherrypy.config.namespaces['scheduler'] = lambda key, value: setattr(cherrypy.scheduler, key, value)



# CherryPy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import tempfile
from collections import namedtuple
from urllib.parse import urlencode

import cherrypy
from cherrypy.lib.sessions import FileSession
from cherrypy.test import helper

from .. import auth  # noqa

User = namedtuple('User', 'id,username')


def checkpassword(username, password):
    return username == 'myuser' and password == 'changeme'


def user_lookup_func(login, user_info):
    if login == 'myuser':
        return login, User(2, login)
    return None


def user_from_key_func(userkey):
    if userkey == 'myuser':
        return User(2, 'myuser')
    return None


@cherrypy.tools.sessions(locking='explicit', storage_class=FileSession)
@cherrypy.tools.auth(
    user_lookup_func=user_lookup_func,
    user_from_key_func=user_from_key_func,
    checkpassword=checkpassword,
)
class Root:

    @cherrypy.expose
    def index(self):
        return "OK"

    @cherrypy.expose()
    def login(self, username=None, password=None):
        if cherrypy.serving.request.method == 'POST' and username and password:
            userobj = cherrypy.tools.auth.login_with_credentials(username, password)
            if userobj:
                raise cherrypy.tools.auth.redirect_to_original_url()
            else:
                return "invalid credentials"
        return "login"


class AuthManagerTest(helper.CPWebCase):
    interactive = False

    @classmethod
    def setup_class(cls):
        cls.tempdir = tempfile.TemporaryDirectory(prefix='cherrypy-foundation-', suffix='-session')
        cls.session_dir = cls.tempdir.name
        super().setup_class()

    @classmethod
    def teardown_class(cls):
        cls.tempdir.cleanup()
        super().teardown_class()

    @classmethod
    def setup_server(cls):
        cherrypy.config.update(
            {
                'tools.sessions.storage_path': cls.session_dir,
            }
        )
        cherrypy.tree.mount(Root(), '/')

    def test_auth_redirect(self):
        # Given unauthenticated user
        # When requesting protected page
        self.getPage('/')
        # Then user is redirected to login page
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/login/')

    def test_auth_login(self):
        # Given unauthenticated user
        # When posting valid login
        self.getPage('/login/', method='POST', body=urlencode({'username': 'myuser', 'password': 'changeme'}))
        # Then user is redirect to index.
        self.assertStatus(303)
        self.assertHeaderItemValue('Location', f'http://{self.HOST}:{self.PORT}/')
        # Then page is accessible
        self.getPage('/', headers=self.cookies)
        self.assertStatus(200)
        self.assertInBody('OK')



# Session timeout tool for cherrypy
# Copyright (C) 2012-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# BFH Science Self-Service Portal
# Copyright (C) 2025-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import time
from contextlib import contextmanager

import cherrypy
import zc.lockfile
from cherrypy.lib import locking
from cherrypy.lib.sessions import FileSession as CPFileSession


# Issue in cherrypy: https://github.com/cherrypy/cherrypy/issues/2065
class FileSession(CPFileSession):
    """
    Override implementation of cherrpy file session to improve file locking.
    """

    def acquire_lock(self, path=None):
        """Acquire an exclusive lock on the currently-loaded session data."""
        if path is None:
            path = self._get_file_path()
        path += self.LOCK_SUFFIX
        checker = locking.LockChecker(self.id, self.lock_timeout)
        while not checker.expired():
            try:
                self.lock = zc.lockfile.LockFile(path)
            except zc.lockfile.LockError:
                # Sleep for 1ms only.
                time.sleep(0.001)
            else:
                break
        self.locked = True
        if self.debug:
            cherrypy.log('Lock acquired.', 'TOOLS.SESSIONS')


@contextmanager
def session_lock():
    """
    Acquire session lock as required. Support re-intrant lock.
    """
    s = cherrypy.serving.session
    if s.locking == 'explicit' and not s.locked:
        s.acquire_lock()
        try:
            yield s
        finally:
            # When explicit, we want to save the session (with also release the lock.)
            if s.locking == 'explicit':
                s.save()
                cherrypy.serving.request._sessionsaved = True
    else:
        yield s



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

project = 'Cherrypy-foundation'
copyright = 'Copyright (C) 2026 IKUS Software'
author = 'Patrik Dufresne'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'myst_parser',
    'sphinx.ext.autosectionlabel',
]

autosectionlabel_prefix_document = True

# Enable anchors for cross references
myst_heading_anchors = 2

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'
# Ref: https://alabaster.readthedocs.io/en/latest/customization.html#theme-options
html_theme_options = {
    "footer_start": ["copyright"],
    "footer_end": ["version"],
    'use_edit_page_button': True,
}
html_context = {
    "gitlab_user": "ikus-soft",
    "gitlab_repo": "cherrypy-foundation",
    "gitlab_version": "main",
    "doc_path": "doc",
}
html_show_sourcelink = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']
